import logging

import traitlets as traits
from nbconvert.preprocessors import Preprocessor
from nbformat.notebooknode import NotebookNode

logger = logging.getLogger("captions")


class LatexCaptions(Preprocessor):
    """ a preprocessor to:

    1. find cells with a ipub.caption meta-tag,
       extract the caption and label to a dict and remove the cell
    2. find cells with the found labels and replace their captions

    """

    add_prefix = traits.Bool(
        False,
        help=("add float type/number prefix to caption " "(from caption_prefix tag)"),
    ).tag(config=True)

    def preprocess(self, nb, resources):

        logger.info("extracting caption cells")

        # extract captions
        final_cells = []
        captions = {}
        for cell in nb.cells:
            if hasattr(cell.metadata, "ipub"):

                if hasattr(cell.metadata.ipub.get("equation", False), "get"):
                    if hasattr(
                        cell.metadata.ipub.equation.get("environment", False),
                        "startswith",
                    ):
                        if cell.metadata.ipub.equation.environment.startswith(
                            "breqn"
                        ):  # noqa: E501
                            if "ipub" not in nb.metadata:
                                nb.metadata["ipub"] = NotebookNode(
                                    {"enable_breqn": True}
                                )
                            else:
                                nb.metadata.ipub["enable_breqn"] = True

                if hasattr(cell.metadata.ipub, "caption"):

                    if cell.cell_type == "markdown":
                        capt = cell.source.split(r"\n")[0]
                        captions[cell.metadata.ipub.caption] = capt
                        continue
                    elif cell.cell_type == "code":
                        if not cell.outputs:
                            pass
                        elif "text/latex" in cell.outputs[0].get("data", {}):
                            capt = cell.outputs[0].data["text/latex"].split(r"\n")[0]
                            captions[cell.metadata.ipub.caption] = capt
                            continue
                        elif "text/plain" in cell.outputs[0].get("data", {}):
                            capt = cell.outputs[0].data["text/plain"].split(r"\n")[0]
                            captions[cell.metadata.ipub.caption] = capt
                            continue

            final_cells.append(cell)
        nb.cells = final_cells

        # replace captions
        for cell in nb.cells:
            if hasattr(cell.metadata, "ipub"):
                for key in cell.metadata.ipub:
                    if hasattr(cell.metadata.ipub[key], "label"):
                        if cell.metadata.ipub[key]["label"] in captions:
                            logger.debug(
                                "replacing caption for: {}".format(
                                    cell.metadata.ipub[key]["label"]
                                )
                            )
                            cell.metadata.ipub[key]["caption"] = captions[
                                cell.metadata.ipub[key]["label"]
                            ]  # noqa: E501

                    # add float type/number prefix to caption, if required
                    if self.add_prefix:
                        if hasattr(cell.metadata.ipub[key], "caption"):
                            if hasattr(cell.metadata.ipub[key], "caption_prefix"):
                                newcaption = (
                                    cell.metadata.ipub[key].caption_prefix
                                    + cell.metadata.ipub[key].caption
                                )
                                cell.metadata.ipub[key].caption = newcaption

        return nb, resources



""" a module to convert between the old (Python script) segment format,
and the new (JSON) one
"""
from typing import Dict, Tuple  # noqa: F401
import os
import ast
import json


def assess_syntax(path):

    with open(path) as file_obj:
        content = file_obj.read()

    syntax_tree = ast.parse(content)

    docstring = ""  # docstring = ast.get_docstring(syntaxTree)
    dct = None
    dtype = None
    for i, child in enumerate(ast.iter_child_nodes(syntax_tree)):
        if i == 0 and isinstance(child, ast.Expr) and isinstance(child.value, ast.Str):
            docstring = child.value.s
        elif isinstance(child, ast.Assign):
            targets = child.targets
            if len(targets) > 1:
                continue
            target = child.targets[0]  # type: ast.Name
            dtype = target.id
            if dtype not in ["tpl_dict", "tplx_dict"]:
                continue
            if not isinstance(child.value, ast.Dict):
                raise ValueError(
                    "expected {} to be of type Dict: {}".format(dtype, child.value)
                )
            dct = child.value
            break

    if dct is None:
        raise IOError("could not find tpl(x)_dict")

    output = {}
    for key, value in zip(dct.keys, dct.values):
        if not isinstance(key, ast.Str):
            raise ValueError("expected {} key to be of type Str: {}".format(dtype, key))
        if not isinstance(value, ast.Str):
            raise ValueError(
                "expected {} value be of type Str: {}".format(dtype, value)
            )
        output[key.s] = value.s

    return {
        "identifier": os.path.splitext(os.path.basename(path))[0],
        "description": docstring,
        "segments": output,
        "$schema": "../../schema/segment.schema.json",
    }


def py_to_json(path, outpath=None):
    output = assess_syntax(path)
    if outpath:
        with open(outpath, "w") as file_obj:
            json.dump(output, file_obj, indent=2)
    return json.dumps(output, indent=2)


if __name__ == "__main__":
    import glob

    _dir_path = os.path.dirname(os.path.realpath(__file__))
    _ext = ".tpl.json"

    for _path in glob.glob(os.path.join(_dir_path, "**", "*.py")):

        _name = os.path.splitext(os.path.basename(_path))[0]
        _folder = os.path.basename(os.path.dirname(_path))
        if _folder == "ipypublish":
            _prefix = "ipy-"
        else:
            _prefix = "std-"

        _outpath = os.path.join(os.path.dirname(_path), _prefix + _name + _ext)

        py_to_json(_path, _outpath)



import os
import sys
from traitlets import Unicode
from ipypublish.postprocessors.base import IPyPostProcessor
from ipypublish.bib2glossary import BibGlossDB


class ConvertBibGloss(IPyPostProcessor):
    """ convert a bibglossary to the required format
    """

    @property
    def allowed_mimetypes(self):
        return None

    @property
    def requires_path(self):
        return True

    @property
    def logger_name(self):
        return "convert-bibgloss"

    encoding = Unicode("utf8", help="the encoding of the input file").tag(config=True)

    resource_key = Unicode(
        "bibglosspath",
        help="the key in the resources dict containing the path to the file",
    ).tag(config=True)

    files_folder = Unicode(
        "_static", help="the path (relative to the main file path) to dump to"
    ).tag(config=True)

    def run_postprocess(self, stream, mimetype, filepath, resources):

        if "bibglosspath" not in resources:
            return stream, filepath, resources

        bibpath = resources["bibglosspath"]

        if not os.path.exists(str(bibpath)):
            self.logger.warning(
                "the bibglossary could not be converted, "
                "since its path does not exist: {}".format(bibpath)
            )
            return stream, filepath, resources

        bibname, extension = os.path.splitext(os.path.basename(bibpath))
        outstr = None
        outext = None

        if extension in [".bib"]:
            if mimetype == "text/restructuredtext":
                pass
            elif mimetype == "text/latex":
                self.logger.info("converting bibglossary to tex")
                bibdb = BibGlossDB()
                bibdb.load_bib(path=str(bibpath), encoding=self.encoding)
                outstr = bibdb.to_latex_string()
                outext = ".tex"

        elif extension in [".tex"]:
            if mimetype == "text/latex":
                pass
            elif mimetype == "text/restructuredtext":
                self.logger.info("converting bibglossary to bibtex")
                bibdb = BibGlossDB()
                bibdb.load_tex(path=str(bibpath), encoding=self.encoding)
                outstr = bibdb.to_bib_string()
                outext = ".bib"

        else:
            self.logger.warning(
                "the bibglossary could not be converted, "
                "since its file extension was not one of: "
                "bib, tex"
            )

        if outstr is None:
            return stream, filepath, resources

        if sys.version_info < (3, 0):
            outstr = unicode(outstr, encoding=self.encoding)  # noqa: F821

        output_folder = filepath.parent.joinpath(self.files_folder)
        if not output_folder.exists():
            output_folder.mkdir(parents=True)

        outfile = output_folder.joinpath(bibname + outext)
        self.logger.info("writing bibglossary: {}".format(outfile))
        with outfile.open("w", encoding=self.encoding) as fh:
            fh.write(outstr)

        self.logger.debug("finished")

        return stream, filepath, resources



import logging

from six import string_types
from traitlets import Bool
from traitlets.config.configurable import Configurable

from ipypublish.utils import handle_error, pathlib

try:
    from shutil import which as exe_exists
except ImportError:
    from distutils.spawn import find_executable as exe_exists  # noqa: F401


class IPyPostProcessor(Configurable):
    """ an abstract class for post-processors
    """

    @property
    def allowed_mimetypes(self):
        """ override in subclasses

        return a list of allowed mime types
        if None, then all are allowed

        Text based mime-types include: text/plain, text/latex,
        text/restructuredtext, text/html, text/x-python, application/json,
        text/markdown, text/asciidoc, text/yaml

        """
        raise NotImplementedError("allowed_mimetypes")

    @property
    def requires_path(self):
        """ override in subclasses

        whether the prostprocessor requires the supplied filepath
        to have an existing parent directory

        if True and filepath is None, will raise an IOError, otherwise,
        will try to make the directory if it doesn't exist

        """
        raise NotImplementedError("requires_path")

    @property
    def logger_name(self):
        """ override in subclass
        """
        return "post-processor"

    @property
    def logger(self):
        return logging.getLogger(self.logger_name)

    skip_mime = Bool(
        True,
        help="if False, raise a TypeError if the mimetype is not allowed, "
        "else return without processing",
    ).tag(config=True)

    def __init__(self, config=None):
        super(IPyPostProcessor, self).__init__(config=config)

    def __call__(self, stream, mimetype, filepath, resources=None):
        """
        See def postprocess() ...
        """
        self.postprocess(stream, mimetype, filepath, resources)

    def postprocess(self, stream, mimetype, filepath, resources=None):
        """ Post-process output.

        Parameters
        ----------
        stream: str
            the main file contents
        mimetype: str
            the mimetype of the file
        filepath: None or str or pathlib.Path
            the path to the output file
            the path does not have to exist, but must be absolute
        resources: None or dict
            a resources dict, output from exporter.from_notebook_node

        Returns
        -------
        stream: str
        filepath: None or str or pathlib.Path

        """

        if (
            self.allowed_mimetypes is not None
            and mimetype not in self.allowed_mimetypes
        ):
            if not self.skip_mime:
                self.handle_error(
                    "the mimetype {0} is not in the allowed list: {1}".format(
                        mimetype, self.allowed_mimetypes
                    ),
                    TypeError,
                )
            else:
                self.logger.debug("skipping incorrect mime type: {}".format(mimetype))
                return stream, filepath, resources

        if self.requires_path and filepath is None:
            self.handle_error(
                "the filepath is None, " "but the post-processor requires a folder",
                IOError,
            )

        if filepath is not None and isinstance(filepath, string_types):
            filepath = pathlib.Path(filepath)

        if self.requires_path:

            if not filepath.is_absolute():
                self.handle_error(
                    "the post-processor requires an absolute folder path", IOError
                )

            if filepath.parent.exists() and not filepath.parent.is_dir():
                self.handle_error(
                    "the filepath's parent is not a folder: {}".format(filepath),
                    TypeError,
                )

            if not filepath.parent.exists():
                filepath.parent.mkdir(parents=True)

        if resources is None:
            resources = {}

        return self.run_postprocess(stream, mimetype, filepath, resources)

    def run_postprocess(self, stream, mimetype, filepath, resources):
        """ should not be called directly
        override in sub-class

        Parameters
        ----------
        stream: str
            the main file contents
        filepath: None or pathlib.Path
            the path to the output file
        resources: dict
            a resources dict, output from exporter.from_notebook_node

        Returns
        -------
        stream: str
        filepath: None or pathlib.Path
        resources: dict

        """
        raise NotImplementedError("run_postprocess")

    def handle_error(self, msg, err_type, raise_msg=None, log_msg=None):
        """ handle error by logging it then raising
        """
        handle_error(msg, err_type, self.logger, raise_msg=raise_msg, log_msg=log_msg)

    def check_exe_exists(self, name, error_msg):
        """ test if an executable exists
        """
        if not exe_exists(name):
            self.handle_error(error_msg, RuntimeError)
        return True


if __name__ == "__main__":

    print(IPyPostProcessor.allowed_mimetypes)
    IPyPostProcessor()("stream", "a")



import os
import shutil
from traitlets import Unicode, List

from ipypublish.postprocessors.base import IPyPostProcessor


class WriteTextFile(IPyPostProcessor):
    """ write the stream to a text based file
    """

    @property
    def allowed_mimetypes(self):
        return (
            "text/latex",
            "text/restructuredtext",
            "text/html",
            "text/x-python",
            "application/json",
            "text/markdown",
        )

    @property
    def requires_path(self):
        return True

    @property
    def logger_name(self):
        return "write-text-file"

    encoding = Unicode("utf8", help="the encoding of the output file").tag(config=True)

    def run_postprocess(self, stream, mimetype, filepath, resources):

        self.logger.info("writing stream to file: {}".format(filepath))
        with filepath.open("w", encoding=self.encoding) as fh:
            fh.write(stream)

        return stream, filepath, resources


class RemoveFolder(IPyPostProcessor):
    """ remove a folder and all its contents
    """

    @property
    def allowed_mimetypes(self):
        return None

    @property
    def requires_path(self):
        return True

    @property
    def logger_name(self):
        return "remove-folder"

    files_folder = Unicode(
        "_static", help="the path (relative to the main file path) to remove"
    ).tag(config=True)

    def run_postprocess(self, stream, mimetype, filepath, resources):

        remove_folder = filepath.parent.joinpath(self.files_folder)
        if remove_folder.exists() and remove_folder.is_dir():
            self.logger.info("removing folder: {0}".format(remove_folder))
            shutil.rmtree(str(remove_folder))

        return stream, filepath, resources


class WriteResourceFiles(IPyPostProcessor):
    """ write content contained in the resources dict to file (as bytes)
    """

    @property
    def allowed_mimetypes(self):
        return None

    @property
    def requires_path(self):
        return True

    @property
    def logger_name(self):
        return "write-resource-files"

    resource_keys = List(
        Unicode(),
        ["outputs"],
        help="the key names in the resources dict that contain files",
    ).tag(config=True)

    # The files already have a relative path
    # files_folder = Unicode(
    #     "_static",
    #     help="the path (relative to the main file path) to write to"
    # ).tag(config=True)

    def run_postprocess(self, stream, mimetype, filepath, resources):

        output_folder = filepath.parent
        if not output_folder.exists():
            output_folder.mkdir(parents=True)

        for key in self.resource_keys:
            if key not in resources:
                continue
            if not hasattr(resources[key], "items"):
                self.handle_error(
                    "the value of resources[{0}] is not a mapping".format(key),
                    TypeError,
                )
            self.logger.info(
                "writing files in resources[{0}] to: {1}".format(key, output_folder)
            )
            for filename, content in resources[key].items():

                outpath = output_folder.joinpath(filename)
                if not outpath.parent.exists():
                    outpath.parent.mkdir(parents=True)

                with outpath.open("wb") as fh:
                    fh.write(content)

        self.logger.debug("finished")

        return stream, filepath, resources


class CopyResourcePaths(IPyPostProcessor):
    """ copy filepaths in the resources dict to another folder
    """

    @property
    def allowed_mimetypes(self):
        return None

    @property
    def requires_path(self):
        return True

    @property
    def logger_name(self):
        return "copy-resource-paths"

    resource_keys = List(
        Unicode(),
        ["external_file_paths"],
        help="the key names in the resources dict that contain filepaths",
    ).tag(config=True)

    files_folder = Unicode(
        "_static", help="the path (relative to the main file path) to copy to"
    ).tag(config=True)

    def run_postprocess(self, stream, mimetype, filepath, resources):

        output_folder = filepath.parent.joinpath(self.files_folder)
        if not output_folder.exists():
            output_folder.mkdir(parents=True)

        for key in self.resource_keys:
            if key not in resources:
                continue
            if not isinstance(resources[key], (list, tuple, set)):
                self.handle_error(
                    "the value of resources[{0}] is not an iterable".format(key),
                    TypeError,
                )
            self.logger.info(
                "copying files in resources[{0}] to: {1}".format(key, output_folder)
            )
            for resfilepath in resources[key]:
                shutil.copyfile(
                    resfilepath,
                    str(output_folder.joinpath(os.path.basename(resfilepath))),
                )

        return stream, filepath, resources



"""
postprocessors that modify the output stream
"""
import re

from ipypublish.postprocessors.base import IPyPostProcessor


class RemoveBlankLines(IPyPostProcessor):
    """ remove multiple lines of blank space """

    @property
    def allowed_mimetypes(self):
        return (
            "text/latex",
            "text/restructuredtext",
            "text/html",
            "text/x-python",
            "application/json",
            "text/markdown",
        )

    @property
    def requires_path(self):
        return False

    @property
    def logger_name(self):
        return "remove-blank-lines"

    def run_postprocess(self, stream, mimetype, filepath, resources):
        stream = re.sub(r"\n\s*\n", "\n\n", stream)
        return stream, filepath, resources


class RemoveTrailingSpace(IPyPostProcessor):
    """ remove trailing whitespace on each line """

    @property
    def allowed_mimetypes(self):
        return (
            "text/latex",
            "text/restructuredtext",
            "text/x-python",
            "application/json",
            "text/markdown",
        )

    @property
    def requires_path(self):
        return False

    @property
    def logger_name(self):
        return "remove-trailing-space"

    def run_postprocess(self, stream, mimetype, filepath, resources):
        stream = "\n".join([l.rstrip() for l in stream.splitlines()])
        return stream, filepath, resources


class FilterOutputFiles(IPyPostProcessor):
    """ filter internal files in resources['outputs'],
    by those that are referenced in the stream """

    @property
    def allowed_mimetypes(self):
        return None

    @property
    def requires_path(self):
        return False

    @property
    def logger_name(self):
        return "filter-output-files"

    def run_postprocess(self, stream, mimetype, filepath, resources):

        if "outputs" in resources:
            for path in list(resources["outputs"].keys()):
                if path not in stream:
                    resources["outputs"].pop(path)

        return stream, filepath, resources


class FixSlideReferences(IPyPostProcessor):
    """ make sure references refer to correct slides """

    @property
    def allowed_mimetypes(self):
        return "text/html"

    @property
    def requires_path(self):
        return False

    @property
    def logger_name(self):
        return "fix-slide-refs"

    def run_postprocess(self, stream, mimetype, filepath, resources):
        if resources and "refslide" in resources:
            for k, (col, row) in resources["refslide"].items():
                stream = stream.replace(
                    "{{id_home_prefix}}{0}".format(k),
                    "#/{0}/{1}{2}".format(col, row, k),
                )
        return stream, filepath, resources



"""
adapted from nbsphinx
"""
import docutils
from docutils import nodes  # noqa E501
from docutils.parsers import rst
from docutils.statemachine import StringList

from ipypublish.sphinx.utils import import_sphinx
from ipypublish.sphinx.notebook.nodes import (
    AdmonitionNode,
    CodeAreaNode,
    FancyOutputNode,
)


class NbAdmonition(rst.Directive):
    """Base class for NbInfo and NbWarning."""

    required_arguments = 0
    optional_arguments = 0
    option_spec = {}
    has_content = True

    def run(self):
        """This is called by the reST parser."""
        node = AdmonitionNode(classes=["admonition", self._class])
        self.state.nested_parse(self.content, self.content_offset, node)
        return [node]


class NbWarning(NbAdmonition):
    """A warning box."""

    _class = "warning"


class NbInfo(NbAdmonition):
    """An information box."""

    _class = "note"


class NBInputToggle(rst.Directive):
    """ a toggle button for nbinput cells """

    _class = "nbinput-toggle-all"
    _default_text = "Toggle Input Cells"

    required_arguments = 0
    optional_arguments = 0
    option_spec = {}
    has_content = True  # button text

    def run(self):
        """This is called by the reST parser."""
        node = nodes.container()
        node["classes"].append(self._class)
        if self.content:
            self.state.nested_parse(self.content, self.content_offset, node)
        else:
            text = (
                self.arguments[0]
                if self.arguments and self.arguments[0]
                else self._default_text
            )
            paragraph = nodes.paragraph(text=text)
            node += paragraph

        return [node]


class NBOutputToggle(NBInputToggle):
    """ a toggle button for nboutput cells """

    _class = "nboutput-toggle-all"
    _default_text = "Toggle Output Cells"


class NbInput(rst.Directive):
    """A notebook input cell with prompt and code area."""

    required_arguments = 0
    optional_arguments = 1  # lexer name
    final_argument_whitespace = False
    option_spec = {
        "execution-count": rst.directives.positive_int,
        "empty-lines-before": rst.directives.nonnegative_int,
        "empty-lines-after": rst.directives.nonnegative_int,
        "no-output": rst.directives.flag,
        "caption": rst.directives.unchanged,
        "name": rst.directives.unchanged,
        "add-toggle": rst.directives.flag,
    }
    has_content = True

    def run(self):
        """This is called by the reST parser."""
        self.state.document["ipysphinx_include_css"] = True
        return _create_nbcell_nodes(self)


class NbOutput(rst.Directive):
    """A notebook output cell with optional prompt."""

    required_arguments = 0
    optional_arguments = 1  # 'rst' or nothing (which means literal text)
    final_argument_whitespace = False
    option_spec = {
        "execution-count": rst.directives.positive_int,
        "more-to-come": rst.directives.flag,
        "empty-lines-before": rst.directives.nonnegative_int,
        "empty-lines-after": rst.directives.nonnegative_int,
        "class": rst.directives.unchanged,
        "add-toggle": rst.directives.flag,
    }
    has_content = True

    def run(self):
        """This is called by the reST parser."""
        self.state.document["ipysphinx_include_css"] = True
        return _create_nbcell_nodes(self)


def _create_nbcell_nodes(directive):
    """Create nodes for an input or output notebook cell."""

    sphinx = import_sphinx()

    language = "none"
    prompt = ""
    fancy_output = False
    execution_count = directive.options.get("execution-count")
    config = directive.state.document.settings.env.config

    if isinstance(directive, NbInput):
        outer_classes = ["nbinput"]
        if "no-output" in directive.options:
            outer_classes.append("nblast")
        inner_classes = ["input_area"]
        if directive.arguments:
            language = directive.arguments[0]
        prompt_template = config.ipysphinx_input_prompt
        if not execution_count:
            execution_count = " "
    elif isinstance(directive, NbOutput):
        outer_classes = ["nboutput"]
        if "more-to-come" not in directive.options:
            outer_classes.append("nblast")
        inner_classes = ["output_area"]
        # 'class' can be 'stderr'
        inner_classes.append(directive.options.get("class", ""))
        prompt_template = config.ipysphinx_output_prompt
        if directive.arguments and directive.arguments[0] in ["rst", "ansi"]:
            fancy_output = True
    else:
        raise AssertionError("directive should be NbInput or NbOutput")

    outer_node = docutils.nodes.container(classes=outer_classes)

    # add prompts
    if config.ipysphinx_show_prompts and execution_count:
        prompt = prompt_template.format(count=execution_count)
        prompt_node = docutils.nodes.literal_block(
            prompt, prompt, language="none", classes=["prompt"]
        )
    elif config.ipysphinx_show_prompts:
        prompt = ""
        prompt_node = docutils.nodes.container(classes=["prompt", "empty"])
    if config.ipysphinx_show_prompts:
        # NB: Prompts are added manually in LaTeX output
        outer_node += sphinx.addnodes.only("", prompt_node, expr="html")

    if fancy_output:
        inner_node = docutils.nodes.container(classes=inner_classes)
        sphinx.util.nodes.nested_parse_with_titles(
            directive.state, directive.content, inner_node
        )
        outtype = directive.arguments[0]
        if outtype == "rst":
            outer_node += FancyOutputNode("", inner_node, prompt=prompt)
        elif outtype == "ansi":
            outer_node += inner_node
        else:
            raise AssertionError(
                "`.. nboutput:: type` should be 'rst' or 'ansi', "
                "not: {}".format(outtype)
            )
    else:
        text = "\n".join(directive.content.data)
        inner_node = docutils.nodes.literal_block(
            text, text, language=language, classes=inner_classes
        )
        codearea_node = CodeAreaNode("", inner_node, prompt=prompt)
        # create a literal text block (e.g. with the code-block directive),
        # that starts or ends with a blank line
        # (see http://stackoverflow.com/q/34050044/)
        for attr in "empty-lines-before", "empty-lines-after":
            value = directive.options.get(attr, 0)
            if value:
                codearea_node[attr] = value

        # add caption and label, see:
        if directive.options.get("caption", False):
            caption = directive.options.get("caption")
            wrapper = container_wrapper(directive, inner_node, caption, inner_classes)
            # add label
            directive.add_name(wrapper)
            outer_node += wrapper
        else:
            outer_node += codearea_node

    if isinstance(directive, NbInput) and (
        config.ipysphinx_input_toggle or "add-toggle" in directive.options
    ):
        directive.state.document["ipysphinx_include_js"] = True
        outer_node += sphinx.addnodes.only(
            "",
            docutils.nodes.container(classes=["toggle-nbinput", "empty"]),
            expr="html",
        )

    if isinstance(directive, NbOutput) and (
        config.ipysphinx_output_toggle or "add-toggle" in directive.options
    ):
        directive.state.document["ipysphinx_include_js"] = True
        outer_node += sphinx.addnodes.only(
            "",
            docutils.nodes.container(classes=["toggle-nboutput", "empty"]),
            expr="html",
        )

    return [outer_node]


def container_wrapper(directive, literal_node, caption, classes):
    """adapted from
    https://github.com/sphinx-doc/sphinx/blob/master/sphinx/directives/code.py
    """
    container_node = docutils.nodes.container(
        "", literal_block=True, classes=classes
    )  # ['literal-block-wrapper']
    parsed = docutils.nodes.Element()
    directive.state.nested_parse(
        StringList([caption], source=""), directive.content_offset, parsed
    )
    if isinstance(parsed[0], docutils.nodes.system_message):
        msg = "Invalid caption: %s" % parsed[0].astext()
        raise ValueError(msg)
    elif isinstance(parsed[0], docutils.nodes.Element):
        caption_node = docutils.nodes.caption(
            parsed[0].rawsource, "", *parsed[0].children
        )
        caption_node.source = literal_node.source
        caption_node.line = literal_node.line
        container_node += caption_node
        container_node += literal_node
        return container_node
    else:
        raise RuntimeError  # never reached



""" serve HTML page
TODO the RevealServer setting should be available at front end
"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function

import os
import signal
import webbrowser

from tornado import web, ioloop, httpserver, log, gen
from tornado.httpclient import AsyncHTTPClient
from traitlets import Bool, Unicode, Int
from ipypublish.postprocessors.base import IPyPostProcessor


class ProxyHandler(web.RequestHandler):
    """handler the proxies requests from a local prefix to a CDN"""

    @gen.coroutine
    def get(self, prefix, url):
        """proxy a request to a CDN"""
        proxy_url = "/".join([self.settings["cdn"], url])
        client = self.settings["client"]
        client.fetch(proxy_url)
        response = yield client.fetch(proxy_url)

        for header in [
            "Content-Type",
            "Cache-Control",
            "Date",
            "Last-Modified",
            "Expires",
        ]:
            if header in response.headers:
                self.set_header(header, response.headers[header])
        self.finish(response.body)


class RevealServer(IPyPostProcessor):
    """Post processor designed to serve files

    Proxies reveal.js requests to a CDN if no local reveal.js is present
    """

    @property
    def allowed_mimetypes(self):
        return "text/html"

    @property
    def requires_path(self):
        return True

    @property
    def logger_name(self):
        return "reveal-server"

    open_in_browser = Bool(
        True, help="Should the browser be opened automatically?"
    ).tag(config=True)

    reveal_cdn = Unicode(
        "https://cdnjs.cloudflare.com/ajax/libs/reveal.js/3.1.0",
        help="""URL for reveal.js CDN.""",
    ).tag(config=True)

    reveal_prefix = Unicode("reveal.js", help="URL prefix for reveal.js").tag(
        config=True
    )

    ip = Unicode("127.0.0.1", help="The IP address to listen on.").tag(config=True)

    port = Int(8000, help="port for the server to listen on.").tag(config=True)

    def run_postprocess(self, stream, mimetype, filepath, resources):
        """Serve the build directory with a webserver."""

        if not filepath.exists():
            self.handle_error(
                "the target file path does not exist: {}".format(filepath), IOError
            )

        # TODO rewrite this as pathlib
        dirname, filename = os.path.split(str(filepath))

        handlers = [
            (r"/(.+)", web.StaticFileHandler, {"path": dirname}),
            (r"/", web.RedirectHandler, {"url": "/%s" % filename}),
        ]

        if "://" in self.reveal_prefix or self.reveal_prefix.startswith("//"):
            # reveal specifically from CDN, nothing to do
            pass
        elif os.path.isdir(os.path.join(dirname, self.reveal_prefix)):
            # reveal prefix exists
            self.logger.info("Serving local %s", self.reveal_prefix)
            self.logger.info("Serving local %s", self.reveal_prefix)
        else:
            self.logger.info(
                "Redirecting %s requests to %s", self.reveal_prefix, self.reveal_cdn
            )
            self.logger.info(
                "Redirecting %s requests to %s", self.reveal_prefix, self.reveal_cdn
            )
            handlers.insert(0, (r"/(%s)/(.*)" % self.reveal_prefix, ProxyHandler))

        app = web.Application(handlers, cdn=self.reveal_cdn, client=AsyncHTTPClient())

        # hook up tornado logging to our self.logger
        log.app_log = self.logger

        http_server = httpserver.HTTPServer(app)

        # find an available port
        port_attempts = list(range(10))
        for port_attempt in port_attempts:
            try:
                url = "http://%s:%i/%s" % (self.ip, self.port, filename)
                self.logger.info("Attempting to serve at %s" % url)
                http_server.listen(self.port, address=self.ip)
                break
            except IOError:
                self.port += 1
        if port_attempt == port_attempts[-1]:
            self.handle_error(
                "no port available to launch slides on, " "try closing some slideshows",
                IOError,
            )

        self.logger.info("Serving your slides at %s" % url)
        self.logger.info("Use Control-C to stop this server")

        # don't let people press ctrl-z, which leaves port open
        def handler(signum, frame):
            self.logger.info("Control-Z pressed, but ignored, use Control-C!")

        signal.signal(signal.SIGTSTP, handler)

        if self.open_in_browser:
            #  2 opens the url in a new tab
            webbrowser.open(url, new=2)

        try:
            ioloop.IOLoop.instance().start()
        except KeyboardInterrupt:
            # dosen't look like line below is necessary
            ioloop.IOLoop.instance().stop()
            self.logger.info("\nInterrupted")

        return stream, filepath, resources



# -*- coding: utf-8 -*-
"""
    test_sphinx
    ~~~~~~~~~~~

    General Sphinx test and check output.
"""
import sys
import pytest
import sphinx

from ipypublish.sphinx.tests import get_test_source_dir

from ipypublish.tests.utils import HTML2JSONParser


@pytest.mark.sphinx(buildername="html", srcdir=get_test_source_dir("bibgloss_basic"))
def test_basic(app, status, warning, get_sphinx_app_output, data_regression):

    app.build()

    assert "build succeeded" in status.getvalue()  # Build succeeded
    warnings = warning.getvalue().strip()
    assert warnings == ""

    output = get_sphinx_app_output(app, buildername="html")

    parser = HTML2JSONParser()
    parser.feed(output)
    if sphinx.version_info >= (2,):
        data_regression.check(parser.parsed, basename="test_basic_v2")
    else:
        data_regression.check(parser.parsed, basename="test_basic_v1")


@pytest.mark.sphinx(buildername="html", srcdir=get_test_source_dir("bibgloss_sortkeys"))
def test_sortkeys(app, status, warning, get_sphinx_app_output, data_regression):

    app.build()

    assert "build succeeded" in status.getvalue()  # Build succeeded
    warnings = warning.getvalue().strip()
    assert warnings == ""

    output = get_sphinx_app_output(app, buildername="html")

    parser = HTML2JSONParser()
    parser.feed(output)
    if sphinx.version_info >= (2,):
        data_regression.check(parser.parsed, basename="test_sortkeys_v2")
    else:
        data_regression.check(parser.parsed, basename="test_sortkeys_v1")


@pytest.mark.sphinx(buildername="html", srcdir=get_test_source_dir("bibgloss_unsorted"))
def test_unsorted(app, status, warning, get_sphinx_app_output, data_regression):

    app.build()

    assert "build succeeded" in status.getvalue()  # Build succeeded
    warnings = warning.getvalue().strip()
    assert warnings == ""

    output = get_sphinx_app_output(app, buildername="html")

    parser = HTML2JSONParser()
    parser.feed(output)
    if sphinx.version_info >= (2,):
        data_regression.check(parser.parsed, basename="test_unsorted_v2")
    else:
        data_regression.check(parser.parsed, basename="test_unsorted_v1")


@pytest.mark.sphinx(
    buildername="html", srcdir=get_test_source_dir("bibgloss_missingref")
)
def test_missingref(app, status, warning, get_sphinx_app_output):

    app.build()

    assert "build succeeded" in status.getvalue()  # Build succeeded
    warnings = warning.getvalue().strip()
    if (
        "could not relabel bibglossary reference [missingkey]" not in warnings
        and "WARNING: citation not found: missingkey" not in warnings  # sphinx < 2
    ):  # sphinx >= 2
        raise AssertionError(
            "should raise warning for missing citation `missingkey`: {}".format(
                warnings
            )
        )


@pytest.mark.sphinx(
    buildername="html", srcdir=get_test_source_dir("bibgloss_duplicatekey")
)
def test_duplicatekey(app, status, warning, get_sphinx_app_output):

    with pytest.raises(KeyError):
        app.build()


@pytest.mark.skipif(
    sys.version_info < (3, 0),
    reason="SyntaxError on import of texsoup/data.py line 135",
)
@pytest.mark.sphinx(buildername="html", srcdir=get_test_source_dir("bibgloss_tex"))
def test_load_tex(app, status, warning, get_sphinx_app_output):

    app.build()

    assert "build succeeded" in status.getvalue()  # Build succeeded
    warnings = warning.getvalue().strip()
    assert warnings == ""



import io
import os
from docutils.parsers import rst
import six

from ipypublish import __version__
from ipypublish.sphinx.utils import import_sphinx

from ipypublish.sphinx.notebook.directives import (
    NbInfo,
    NbInput,
    NbOutput,
    NbWarning,
    AdmonitionNode,
    NBInputToggle,
    NBOutputToggle,
    CodeAreaNode,
    FancyOutputNode,
)
from ipypublish.sphinx.notebook.transforms import (
    CreateDomainObjectLabels,
    CreateSectionLabels,
    RewriteLocalLinks,
)
from ipypublish.sphinx.notebook.parser import NBParser

try:
    from sphinx.application import Sphinx  # noqa: F401
except ImportError:
    pass


def setup(app):
    # type: (Sphinx) -> dict
    """Initialize Sphinx extension.


    Notes
    -----

    TODO better latex output
    but not really interested in this as it would be duplication of effort,
    and if much better todo ipynb -> tex, rather than ipynb -> rst -> tex
    TODO handling of svg in latex
    ipypublish sets latex to output svg rather than pdf, so we don't have to
    split output into '.. only html/latex', which is an issue its something
    with a label (duplication error), however,
    this requires sphinx.ext.imgconverter to work

    """
    # delayed import of sphinx
    sphinx = import_sphinx()

    transforms = app.registry.get_transforms()

    def add_transform(transform, post=False):
        if transform not in transforms:
            if post:
                app.add_post_transform(transform)
            else:
                app.add_transform(transform)

    app.add_source_parser(NBParser)

    associate_single_extension(app, ".ipynb")

    # config for export config
    app.add_config_value(
        "ipysphinx_export_config", "sphinx_ipypublish_all.ext", rebuild="env"
    )

    # config for contolling conversion process
    # where to dump internal images, etc of the notebook
    app.add_config_value("ipysphinx_folder_suffix", "_nbfiles", rebuild="env")
    # whether to raise error if nb_name.rst already exists
    app.add_config_value("ipysphinx_overwrite_existing", False, rebuild="env")
    # additional folders containing conversion files
    app.add_config_value("ipysphinx_config_folders", (), rebuild="env")

    # config for cell prompts
    app.add_config_value("ipysphinx_show_prompts", False, rebuild="env")
    app.add_config_value("ipysphinx_input_prompt", "[{count}]:", rebuild="env")
    app.add_config_value("ipysphinx_output_prompt", "[{count}]:", rebuild="env")

    # config for cell toggling
    app.add_config_value("ipysphinx_input_toggle", False, rebuild="env")
    app.add_config_value("ipysphinx_output_toggle", False, rebuild="env")

    # config for html css
    app.add_config_value("ipysphinx_responsive_width", "540px", rebuild="html")
    app.add_config_value("ipysphinx_prompt_width", None, rebuild="html")
    # setup html style
    app.connect("config-inited", set_css_prompts)
    app.connect("html-page-context", html_add_css)
    # add javascript
    app.connect("html-page-context", html_add_javascript)
    # config for displaying ipywidgets
    app.add_config_value("ipysphinx_always_add_jsurls", False, rebuild="html")
    app.add_config_value("ipysphinx_require_jsurl", None, rebuild="html")
    app.add_config_value("ipysphinx_requirejs_options", None, rebuild="html")
    app.connect("env-updated", add_require_js_path)
    app.add_config_value("ipysphinx_widgets_jsurl", None, rebuild="html")
    app.add_config_value("ipysphinx_widgetsjs_options", {}, rebuild="html")
    app.connect("env-updated", add_ipywidgets_js_path)
    app.connect("env-purge-doc", discard_document_variables)

    # config for additions to the output rst (per file)
    # these strings are processed by the exporters jinja template
    app.add_config_value("ipysphinx_prolog", None, rebuild="env")
    app.add_config_value("ipysphinx_epilog", None, rebuild="env")

    # map additional file extensions to pre-converters
    # NB: jupytext is already a default for .Rmd
    app.add_config_value("ipysphinx_preconverters", {}, rebuild="env")
    app.connect("config-inited", associate_extensions)

    # add the main directives
    app.add_directive("nbinput", NbInput)
    app.add_directive("nboutput", NbOutput)
    app.add_directive("nbinfo", NbInfo)
    app.add_directive("nbwarning", NbWarning)
    app.add_directive("nbinput-toggle-all", NBInputToggle)
    app.add_directive("nboutput-toggle-all", NBOutputToggle)

    # add docutils nodes and visit/depart wraps
    app.add_node(
        CodeAreaNode,
        html=(lambda self, node: None, depart_codearea_html),
        #  latex=(
        #      lambda self, node: self.pushbody([]),  # used in depart
        #      lambda self, node: None,
        #  )
    )
    app.add_node(
        FancyOutputNode,
        html=(lambda self, node: None, lambda self, node: None),
        #  latex=(
        #      lambda self, node: None,
        #      lambda self, node: None,
        #  )
    )
    app.add_node(
        AdmonitionNode,
        html=(visit_admonition_html, lambda self, node: self.body.append("</div>\n")),
        #  latex=(
        #      lambda self, node:
        #      self.body.append(
        #          '\n\\begin{{sphinxadmonition}}{{{class}}}'
        #          '{{}}\\unskip'.format(node['classes'][1])),
        #      lambda self, node:
        #      self.body.append('\\end{sphinxadmonition}\n')
        #  )
    )

    # add transformations
    add_transform(CreateSectionLabels)
    add_transform(CreateDomainObjectLabels)
    add_transform(RewriteLocalLinks)

    # Work-around until https://github.com/sphinx-doc/sphinx/pull/5504 is done:
    mathjax_config = app.config._raw_config.setdefault("mathjax_config", {})
    mathjax_config.setdefault(
        "tex2jax",
        {
            "inlineMath": [["$", "$"], ["\\(", "\\)"]],
            "processEscapes": True,
            "ignoreClass": "document",
            "processClass": "math|output_area",
        },
    )

    # Make docutils' "code" directive (generated by markdown2rst/pandoc)
    # behave like Sphinx's "code-block",
    # see https://github.com/sphinx-doc/sphinx/issues/2155:
    rst.directives.register_directive("code", sphinx.directives.code.CodeBlock)

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
        "env_version": 1,
    }


def associate_extensions(app, config):
    for suffix in config.ipysphinx_preconverters:
        associate_single_extension(app, suffix, config_value="ipysphinx_preconverters")


def associate_single_extension(app, extension, suffix="jupyter_notebook"):
    # type: (Sphinx, str) -> None
    """Associate a file extension with the NBParser."""
    if not isinstance(extension, six.string_types):
        raise AssertionError("extension is not a string: {}".format(extension))
    if not extension.startswith("."):
        raise AssertionError("extension should start with a '.': {}".format(extension))

    sphinx = import_sphinx()

    app.add_source_suffix(extension, suffix)
    sphinx.util.logging.getLogger("nbparser").info(
        "ipypublish: associated {} with NBParser".format(extension), color="green"
    )


def visit_admonition_html(self, node):
    """add openeing div and set classes  """
    self.body.append(self.starttag(node, "div"))
    if len(node.children) >= 2:
        node[0]["classes"].append("admonition-title")
        html_theme = self.settings.env.config.html_theme
        if html_theme in ("sphinx_rtd_theme", "julia"):
            node.children[0]["classes"].extend(["fa", "fa-exclamation-circle"])


def depart_codearea_html(self, node):
    """Add empty lines before and after the code."""
    text = self.body[-1]
    text = text.replace("<pre>", "<pre>\n" + "\n" * node.get("empty-lines-before", 0))
    text = text.replace("</pre>", "\n" * node.get("empty-lines-after", 0) + "</pre>")
    self.body[-1] = text


def read_css(name):
    """Return the contents of a CSS file resource."""
    # TODO use importlib_resources to access CSS file content
    folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), "css")
    with io.open(os.path.join(folder, name + ".css")) as fobj:
        content = fobj.read()

    return content


def set_css_prompts(app, config):
    """Set default value for CSS prompt width (optimized for two-digit numbers)."""
    if config.ipysphinx_prompt_width is None:
        config.ipysphinx_prompt_width = {
            "agogo": "4ex",
            "alabaster": "5ex",
            "better": "5ex",
            "classic": "4ex",
            "cloud": "5ex",
            "dotted": "5ex",
            "guzzle_sphinx_theme": "6ex",
            "haiku": "4ex",
            "julia": "5ex",
            "maisie_sphinx_theme": "6ex",
            "nature": "5ex",
            "pangeo": "5ex",
            "pyramid": "5ex",
            "redcloud": "5ex",
            "sizzle": "5.5ex",
            "sphinx_py3doc_enhanced_theme": "6ex",
            "sphinx_pyviz_theme": "5.5ex",
            "sphinx_rtd_theme": "5ex",
            "sphinx_typlog_theme": "5.5ex",
            "traditional": "4ex",
        }.get(config.html_theme, "7ex")


def html_add_css(app, pagename, templatename, context, doctree):
    """Add CSS string to HTML pages that contain code cells."""
    style = ""
    if doctree and doctree.get("ipysphinx_include_css"):
        style += read_css("nb_cells") % app.config
    if doctree and app.config.html_theme in ("sphinx_rtd_theme", "julia"):
        style += read_css("sphinx_rtd_theme")
    if doctree and app.config.html_theme in ("cloud", "redcloud"):
        style += read_css("cloud_theme")
    if style:
        context["body"] = "\n<style>" + style + "</style>\n" + context["body"]


def copy_javascript(name):
    """Return the contents of javascript resource file."""
    # TODO use importlib_resources to access javascript file content
    folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), "js")
    with open(os.path.join(folder, name + ".js")) as fobj:
        content = fobj.read()

    return content


def html_add_javascript(app, pagename, templatename, context, doctree):
    """Add JavaScript string to HTML pages that contain code cells."""
    if doctree and doctree.get("ipysphinx_include_js"):
        code = copy_javascript("toggle_code")
        context["body"] = "\n<script>" + code + "</script>\n" + context["body"]
        code = copy_javascript("toggle_output")
        context["body"] = "\n<script>" + code + "</script>\n" + context["body"]


def add_require_js_path(app, env):
    """Insert the require javascript package url, to pages created from notebooks."""
    config = app.config
    if not (
        getattr(env, "ipysphinx_created_from_nb", set())
        or config.ipysphinx_always_add_jsurls
    ):
        return
    if config.ipysphinx_require_jsurl is None:
        requirejs_path = (
            "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js"
        )
    else:
        requirejs_path = config.ipysphinx_require_jsurl
    if config.ipysphinx_requirejs_options is None:
        requirejs_options = {
            "integrity": "sha256-Ae2Vz/4ePdIu6ZyI/5ZGsYnb+m0JlOmKPjt6XZ9JJkA=",
            "crossorigin": "anonymous",
        }
    else:
        requirejs_options = config.ipysphinx_requirejs_options

    app.add_js_file(requirejs_path, **requirejs_options)


def add_ipywidgets_js_path(app, env):
    """Insert the ipywidgets javascript url, to pages created from notebooks containing widgets."""
    if not (
        getattr(env, "ipysphinx_widgets", set())
        or app.config.ipysphinx_always_add_jsurls
    ):
        return
    sphinx = import_sphinx()
    widgets_path = None
    if app.config.ipysphinx_widgets_jsurl is None:
        try:
            from ipywidgets.embed import DEFAULT_EMBED_REQUIREJS_URL
        except ImportError:
            logger = sphinx.util.logging.getLogger(__name__)
            logger.warning(
                "ipysphinx_widgets_jsurl not given and ipywidgets module unavailable"
            )
        else:
            widgets_path = DEFAULT_EMBED_REQUIREJS_URL
    else:
        widgets_path = app.config.ipysphinx_widgets_jsurl
    if widgets_path is not None:
        app.add_js_file(widgets_path, **app.config.ipysphinx_widgetsjs_options)


def discard_document_variables(app, env, docname):
    """Discard any document specific variables."""
    # Discard any ipywidgets gathered from the current notebook.
    try:
        env.ipysphinx_widgets.discard(docname)
    except AttributeError:
        pass
    try:
        env.ipysphinx_created_from_nb.discard(docname)
    except AttributeError:
        pass



import os
import logging
from six import string_types
from ipypublish.utils import read_file_from_directory

logger = logging.getLogger("make-conf")


def make_conf(overwrite=False, **kwargs):
    """ create the contents of a sphinx config.py
    with default values compliant with ipypublish.

    Parameters
    ----------
    overwrite: bool
        if True, default arguments are overwritten by key-word arguments
    kwargs:
        key-word arguments are included as <key> = <val>
        note docstring is a special key, inserted at the top of the file

    Returns
    -------
    conf_content: str

    """

    # load local config.yaml
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)))
    data = read_file_from_directory(
        path, "config.yaml", "sphinx config", logger, interp_ext=True
    )

    conf_str = []

    docstring = None
    if "docstring" in data:
        docstring = data.pop("docstring")

    if "docstring" in kwargs:
        kwdocstring = data.pop("docstring")
        if overwrite or not docstring:
            docstring = kwdocstring

    if docstring:
        conf_str.append('"""')
        conf_str.append(str(docstring))
        conf_str.append('"""')

    for key, val in data.items():
        if key in kwargs and overwrite:
            continue
        conf_str.append("")
        if isinstance(val, string_types):
            val = '"{}"'.format(val)
        conf_str.append("{0} = {1}".format(key, val))

    for key, val in kwargs.items():
        if key in data and not overwrite:
            continue
        conf_str.append("")
        if isinstance(val, string_types):
            val = '"{}"'.format(val)
        conf_str.append("{0} = {1}".format(key, val))

    return "\n".join(conf_str) + "\n"


def make_index(
    toc_files,
    toc_depth=3,
    toc_title="Table of Contents",
    toc_numbered=True,
    toc_glob=False,
    header=None,
    prolog=None,
    epilog=None,
):
    """ make an index file, containing a toc tree

    Parameters
    ----------
    toc_files: list[str]
        list of file paths (relative to the index)
        to be included in the toc tree
    toc_depth=3: int
        depth of toc tree
    toc_title: str
        title of toc tree
    toc_numbered=True: bool
        number sections and figures, tables, etc
    header: None or str
    prolog: None or str
    epilog: None or str

    Returns
    -------
    index_content: str

    """
    if not toc_files:
        raise AssertionError("there must be at lease on toc file")

    index_str = []

    if header:
        index_str.append(str(header))
        index_str.append("=" * len(str(header)))
        index_str.append("")

    if prolog:
        index_str.append(str(prolog))
        index_str.append("")

    index_str.append(".. toctree::")
    index_str.append("   :includehidden:")
    index_str.append("   :maxdepth: {}".format(toc_depth))
    if toc_numbered:
        index_str.append("   :numbered:")
    if toc_glob:
        index_str.append("   :glob:")
    index_str.append("   :caption: {}:".format(toc_title))
    index_str.append("")

    for toc_file in toc_files:
        index_str.append("   " + os.path.splitext(toc_file)[0].lstrip())

    if epilog:
        index_str.append("")
        index_str.append(str(epilog))

    return "\n".join(index_str) + "\n"


if __name__ == "__main__":

    print(
        make_conf(
            project="ipypublish",
            author="Chris Sewell",
            description=(
                "Create quality publication and presentation"
                "directly from Jupyter Notebook(s)"
            ),
        )
    )

    print(make_index(["path/to/file.rst"], header="Header"))



import copy
import logging
import re

import traitlets as traits
from nbconvert.preprocessors import Preprocessor
from nbformat.notebooknode import NotebookNode


class FinalCells(object):
    """ a class that stores cells
    """

    def __init__(self, header_slide):
        self.cells = []
        if header_slide:
            self.horizontalbreak_after = "horizontalbreak_after_plusvertical"
        else:
            self.horizontalbreak_after = "horizontalbreak_after"

    def mkdcell(self, source, metadata, slidetype):
        meta = copy.deepcopy(metadata)
        meta.ipyslides = slidetype
        self.append(
            NotebookNode(
                {"cell_type": "markdown", "source": "\n".join(source), "metadata": meta}
            )
        )

    def append(self, cell):
        last = self.last()
        if not last:
            pass
        elif cell.metadata.ipyslides == "verticalbreak_after":
            pass  # last.metadata.ipyslides = 'verticalbreak_above'
        elif cell.metadata.ipyslides == self.horizontalbreak_after:
            # if last.metadata.ipyslides == 'before_header':
            #    last.metadata.ipyslides == 'between_headers'
            if not last.metadata.ipyslides == self.horizontalbreak_after:
                last.metadata.ipyslides = "horizontalbreak_before"
            else:
                last.metadata.ipyslides = "horizontalbreak_after_novertical"
        self.cells.append(cell)

    def first(self):
        for cell in self.cells:
            if cell.metadata.ipyslides not in ["skip", "notes"]:
                return cell
        return False

    def last(self):
        for cell in reversed(self.cells):
            if cell.metadata.ipyslides not in ["skip", "notes"]:
                return cell
        return False

    def finalize(self):
        if not self.first():
            return False
        if self.first().metadata.ipyslides == "normal":
            self.first().metadata.ipyslides = "first_cell"
        if self.last().metadata.ipyslides == "normal":
            self.last().metadata.ipyslides = "last_cell"
        return True


def is_header(line, max_level):
    """if max_level is 0 assumes all headers ok

    Examples
    --------
    >>> is_header("abc",0)
    False
    >>> is_header("#",0)
    False
    >>> is_header("# title",0)
    True
    >>> is_header("### title",3)
    True
    >>> is_header("### title",2)
    False

    """
    if max_level:
        return len(re.findall("^#{{1,{0}}} .+".format(max_level), line)) > 0
    else:
        return len(re.findall("^#+ .+", line)) > 0


def header_level(line):
    """

    Examples
    --------
    >>> header_level('# title')
    1
    >>> header_level('### title')
    3
    """
    i = 0
    title = line + "e"
    while title[0] == "#":
        i += 1
        title = title[1:]
    return i


def number_title(line, current_levels):
    """

    Examples
    --------
    >>> number_title("# title",[])
    ('# 1. title', [1])
    >>> number_title("## title",[])
    ('## 1.1. title', [1, 1])
    >>> number_title("# title",[1,1])
    ('# 2. title', [2])
    >>> number_title("## title",[2,1])
    ('## 2.2. title', [2, 2])
    >>> number_title("### title a#bc",[2])
    ('### 2.1.1. title a#bc', [2, 1, 1])
    >>> number_title("### title a#bc",[2,1,2,3])
    ('### 2.1.3. title a#bc', [2, 1, 3])
    """
    level = header_level(line)
    if not level > 0:
        raise ValueError("level must be > 0: {}".format(level))
    if len(current_levels) < level:
        while len(current_levels) < level:
            current_levels.append(1)
    else:
        current_levels = current_levels[:level]
        current_levels[-1] += 1
    hashes, title = line.split(" ", 1)
    numbers = ".".join([str(i) for i in current_levels]) + "."
    new = " ".join([hashes, numbers, title])
    return new, current_levels


class MarkdownSlides(Preprocessor):
    """ a preprocessor to setup the notebook as an ipyslideshow,
    according to a set of rules

    - markdown cells containaing # headers are broken into individual cells
    - any cells where ipub.ignore=True is set to 'skip'
    - any code cells with no other ipub tags are set to 'skip'
    - any header level >= column_level starts a new column
    - else, any header level >= row_level starts a new row
    - if max_cells is not 0, then breaks to a new row after <max_cells> cells

    """

    column_level = traits.Integer(
        1, min=0, help="maximum header level for new columns (0 indicates no maximum)"
    ).tag(config=True)
    row_level = traits.Integer(
        0, min=0, help="maximum header level for new rows (0 indicates no maximum)"
    ).tag(config=True)
    header_slide = traits.Bool(
        False,
        help=("if True, make the first header in a " "column appear on its own slide"),
    ).tag(config=True)
    max_cells = traits.Integer(
        0, min=0, help="maximum number of nb cells per slide (0 indicates no maximum)"
    ).tag(config=True)
    autonumbering = traits.Bool(
        False, help="append section numbering to titles, e.g. 1.1.1 Title"
    ).tag(config=True)

    def preprocess(self, nb, resources):

        logging.info("creating slides based on markdown and existing slide tags")
        latexdoc_tags = ["code", "error", "table", "equation", "figure", "text"]
        # break up titles
        cells_in_slide = 0
        final_cells = FinalCells(self.header_slide)

        header_levels = []
        try:
            base_numbering = nb.metadata.toc.base_numbering
            header_levels = list(map(lambda x: int(x), base_numbering.split(".")))
            header_levels[0] -= 1
            logging.debug("base_numbering = " + base_numbering)
            logging.debug("header_levels = " + str(header_levels))
        except ValueError:
            logging.warning("Invalid toc.base_numbering in notebook metadata")
        except AttributeError:
            logging.debug("No toc.base_numbering in notebook metadata; starting at 1")

        for i, cell in enumerate(nb.cells):

            # Make sure every cell has an ipub meta tag
            cell.metadata.ipub = cell.metadata.get("ipub", NotebookNode())

            if cell.metadata.ipub.get("ignore", False):
                cell.metadata.ipyslides = "skip"
                final_cells.append(cell)
                continue

            if cell.metadata.ipub.get("slide", False) == "notes":
                cell.metadata.ipyslides = "notes"
                final_cells.append(cell)
                continue

            if not cell.cell_type == "markdown":
                # TODO this doesn't test if the data is actually available
                # to be output
                if not any(
                    [cell.metadata.ipub.get(typ, False) for typ in latexdoc_tags]
                ):
                    cell.metadata.ipyslides = "skip"
                    final_cells.append(cell)
                    continue

                if cells_in_slide > self.max_cells and self.max_cells:
                    cell.metadata.ipyslides = "verticalbreak_after"
                    cells_in_slide = 1
                elif cell.metadata.ipub.get("slide", False) == "new":
                    cell.metadata.ipyslides = "verticalbreak_after"
                    cells_in_slide = 1
                else:
                    cell.metadata.ipyslides = "normal"
                    cells_in_slide += 1
                final_cells.append(cell)
                continue

            nonheader_lines = []
            for line in cell.source.split("\n"):

                if is_header(line, 0) and self.autonumbering:
                    line, header_levels = number_title(line, header_levels[:])

                if is_header(line, self.column_level):
                    if nonheader_lines and cell.metadata.ipub.get("slide", False):
                        if (
                            cells_in_slide > self.max_cells and self.max_cells
                        ) or cell.metadata.ipub.slide == "new":
                            final_cells.mkdcell(
                                nonheader_lines, cell.metadata, "verticalbreak_after"
                            )
                            cells_in_slide = 1
                        else:
                            cells_in_slide += 1
                            final_cells.mkdcell(
                                nonheader_lines, cell.metadata, "normal"
                            )
                        # current_lines = []

                    if self.header_slide:
                        final_cells.mkdcell(
                            [line], cell.metadata, "horizontalbreak_after_plusvertical"
                        )
                    else:
                        final_cells.mkdcell(
                            [line], cell.metadata, "horizontalbreak_after"
                        )
                    cells_in_slide = 1

                elif is_header(line, self.row_level):
                    if nonheader_lines and cell.metadata.ipub.get("slide", False):
                        if (
                            cells_in_slide > self.max_cells and self.max_cells
                        ) or cell.metadata.ipub.slide == "new":
                            final_cells.mkdcell(
                                nonheader_lines, cell.metadata, "verticalbreak_after"
                            )
                            cells_in_slide = 1
                        else:
                            cells_in_slide += 1
                            final_cells.mkdcell(
                                nonheader_lines, cell.metadata, "normal"
                            )
                        # current_lines = []

                    final_cells.mkdcell([line], cell.metadata, "verticalbreak_after")
                    cells_in_slide = 1
                else:
                    nonheader_lines.append(line)

            if nonheader_lines and cell.metadata.ipub.get("slide", False):
                if (
                    cells_in_slide > self.max_cells and self.max_cells
                ) or cell.metadata.ipub.slide == "new":
                    final_cells.mkdcell(
                        nonheader_lines, cell.metadata, "verticalbreak_after"
                    )
                    cells_in_slide = 1
                else:
                    cells_in_slide += 1
                    final_cells.mkdcell(nonheader_lines, cell.metadata, "normal")

        if not final_cells.finalize():
            logging.warning("no cells available for slideshow")
        nb.cells = final_cells.cells

        return nb, resources



import pytest


@pytest.mark.ipynb("basic_nb")
def test_basic_latex(ipynb_app):
    ipynb_app.run({"conversion": "latex_ipypublish_main"})
    ipynb_app.assert_converted_exists()


@pytest.mark.ipynb("basic_nb")
def test_basic_html(ipynb_app):
    ipynb_app.run({"conversion": "html_ipypublish_main"})
    ipynb_app.assert_converted_exists()


@pytest.mark.ipynb("basic_nb")
def test_basic_slides(ipynb_app):
    ipynb_app.run(
        {
            "conversion": "slides_ipypublish_main",
            "default_pporder_kwargs": {"slides": True},
        }
    )
    ipynb_app.assert_converted_exists()


@pytest.mark.requires_latexmk
@pytest.mark.ipynb("basic_nb")
def test_basic_pdf(ipynb_app):
    ipynb_app.run(
        {
            "conversion": "latex_ipypublish_main",
            "default_pporder_kwargs": {"create_pdf": True},
        }
    )
    ipynb_app.assert_converted_exists()
    ipynb_app.assert_converted_exists(extension=".pdf")



from ipypublish.filters_pandoc.main import jinja_filter
from ipypublish.filters_pandoc.utils import create_ipub_meta


def test_basic():

    out_str = jinja_filter("a", "rst", {}, {})
    assert out_str == "a"


def test_reference():

    out_str = jinja_filter("@label", "rst", {}, {})
    assert out_str == ":cite:`label`"


def test_reference_prefix():

    out_str = jinja_filter("+@label", "rst", {}, {})
    assert out_str == ":numref:`label`"


def test_option_in_nb_meta():

    out_str = jinja_filter(
        "+@label", "rst", create_ipub_meta({"use_numref": False}), {}
    )
    assert out_str == ":ref:`label`"


def test_option_in_cell_meta():

    out_str = jinja_filter(
        "+@label",
        "rst",
        create_ipub_meta({"use_numref": False}),
        create_ipub_meta({"use_numref": True}),
    )
    assert out_str == ":numref:`label`"


def test_option_in_top_matter():
    # TODO create ipub yaml from IPUB_META_ROUTE

    in_str = "\n".join(
        ["---", "ipub:", "  pandoc:", "    use_numref: true", "", "...", "", "+@label"]
    )

    out_str = jinja_filter(in_str, "rst", create_ipub_meta({"use_numref": False}), {})
    assert out_str == ":numref:`label`"


def test_at_notation_false():

    out_str = jinja_filter(
        "+@label", "rst", create_ipub_meta({"at_notation": False}), {}
    )
    assert out_str == "+ :cite:`label`"


def test_remove_filter():

    out_str = jinja_filter(
        "+@label", "rst", create_ipub_meta({"apply_filters": False}), {}
    )
    assert out_str == "+@label"


def test_complex_equation():

    in_source = [
        "$$\\begin{equation*}\n",
        "f(x) = \\left\\{\n",
        "\\begin{array}{ll}\n",
        "\\; x \\qquad x \\geq 0 \\\\\n",
        "\\; 0 \\qquad else\n",
        "\\end{array}\n",
        "\\right.\n",
        "\\end{equation*}$$",
    ]

    out_string = jinja_filter("".join(in_source), "rst", create_ipub_meta({}), {})
    expected = [
        ".. math::",
        "   :nowrap:",
        "",
        "   \\begin{equation*}",
        "   f(x) = \\left\\{",
        "   \\begin{array}{ll}",
        "   \\; x \\qquad x \\geq 0 \\\\",
        "   \\; 0 \\qquad else",
        "   \\end{array}",
        "   \\right.",
        "   \\end{equation*}",
    ]
    assert out_string.strip() == "\n".join(expected)



import os
import glob
import importlib
import logging

from six import string_types
from jinja2 import DictLoader
import jsonschema
import nbconvert  # noqa: F401

from ipypublish.utils import (
    pathlib,
    handle_error,
    get_module_path,
    read_file_from_directory,
    read_file_from_module,
)
from ipypublish import export_plugins
from ipypublish import schema
from ipypublish.templates.create_template import create_template

_TEMPLATE_KEY = "new_template"
_EXPORT_SCHEMA_FILE = "export_config.schema.json"
_EXPORT_SCHEMA = None

logger = logging.getLogger("configuration")


def get_export_config_path(export_key, config_folder_paths=()):
    # type (string, Tuple[str]) -> Union[string, None]
    """we search for a plugin name, which matches the supplied plugin name
    """
    for name, jsonpath in iter_all_export_paths(config_folder_paths):
        if name == export_key:
            return pathlib.Path(jsonpath)
    return None


def iter_all_export_paths(config_folder_paths=(), regex="*.json"):
    """we iterate through all json files in the
    supplied plugin_folder_paths, and then in the `export_plugins` folder
    """
    for plugin_folder_path in config_folder_paths:
        for jsonpath in glob.glob(os.path.join(plugin_folder_path, regex)):
            name = os.path.splitext(os.path.basename(jsonpath))[0]
            yield name, pathlib.Path(jsonpath)

    module_path = get_module_path(export_plugins)
    for jsonpath in glob.glob(os.path.join(str(module_path), regex)):
        name = os.path.splitext(os.path.basename(jsonpath))[0]
        yield name, pathlib.Path(jsonpath)


def load_export_config(export_config_path):
    """load the export configuration"""
    if isinstance(export_config_path, string_types):
        export_config_path = pathlib.Path(export_config_path)

    data = read_file_from_directory(
        export_config_path.parent,
        export_config_path.name,
        "export configuration",
        logger,
        interp_ext=True,
    )

    # validate against schema
    global _EXPORT_SCHEMA
    if _EXPORT_SCHEMA is None:
        # lazy load schema once
        _EXPORT_SCHEMA = read_file_from_directory(
            get_module_path(schema),
            _EXPORT_SCHEMA_FILE,
            "export configuration schema",
            logger,
            interp_ext=True,
        )
    try:
        jsonschema.validate(data, _EXPORT_SCHEMA)
    except jsonschema.ValidationError as err:
        handle_error(
            "validation of export config {} failed against {}: {}".format(
                export_config_path, _EXPORT_SCHEMA_FILE, err.message
            ),
            jsonschema.ValidationError,
            logger=logger,
        )

    return data


def iter_all_export_infos(config_folder_paths=(), regex="*.json", get_mime=False):
    """iterate through all export configuration and yield a dict of info"""
    for name, path in iter_all_export_paths(config_folder_paths, regex):
        data = load_export_config(path)

        info = dict(
            [
                ("key", str(name)),
                ("class", data["exporter"]["class"]),
                ("path", str(path)),
                ("description", data["description"]),
            ]
        )

        if get_mime:
            info["mime_type"] = create_exporter_cls(
                data["exporter"]["class"]
            ).output_mimetype

        yield info


def create_exporter_cls(class_str):
    # type: (str) -> nbconvert.exporters.Exporter
    """dynamically load export class"""
    export_class_path = class_str.split(".")
    module_path = ".".join(export_class_path[0:-1])
    class_name = export_class_path[-1]
    try:
        export_module = importlib.import_module(module_path)
    except ModuleNotFoundError:  # noqa: F821
        handle_error(
            "module {} containing exporter class {} not found".format(
                module_path, class_name
            ),
            ModuleNotFoundError,
            logger=logger,
        )  # noqa: F821
    if hasattr(export_module, class_name):
        export_class = getattr(export_module, class_name)
    else:
        handle_error(
            "module {} does not contain class {}".format(module_path, class_name),
            ImportError,
            logger=logger,
        )

    return export_class


def get_export_extension(export_config_path):
    """return the file extension of the exporter class"""
    data = load_export_config(export_config_path)
    exporter_cls = create_exporter_cls(data["exporter"]["class"])
    return exporter_cls.file_extension


def str_to_jinja(template_str, template_key="jinja_template"):
    return DictLoader({template_key: template_str})


def load_template(template_key, template_dict):

    if template_dict is None:
        return None

    if "directory" in template_dict["outline"]:
        outline_template = read_file_from_directory(
            template_dict["outline"]["directory"],
            template_dict["outline"]["file"],
            "template outline",
            logger,
            interp_ext=False,
        )
        outline_name = "{0}/{1}".format(
            template_dict["outline"]["directory"], template_dict["outline"]["file"]
        )
    else:
        outline_template = read_file_from_module(
            template_dict["outline"]["module"],
            template_dict["outline"]["file"],
            "template outline",
            logger,
            interp_ext=False,
        )
        outline_name = "{0}/{1}".format(
            template_dict["outline"]["module"], template_dict["outline"]["file"]
        )

    segments = []
    for snum, segment in enumerate(template_dict.get("segments", [])):

        if "file" not in segment:
            handle_error("'file' expected in segment {}".format(snum), KeyError, logger)

        if "directory" in segment:
            seg_data = read_file_from_directory(
                segment["directory"],
                segment["file"],
                "template segment",
                logger,
                interp_ext=True,
            )
        elif "module" in segment:
            seg_data = read_file_from_module(
                segment["module"],
                segment["file"],
                "template segment",
                logger,
                interp_ext=True,
            )
        else:
            handle_error(
                "'directory' or 'module' expected in segment {}".format(snum),
                KeyError,
                logger,
            )

        segments.append(seg_data)

    template_str = create_template(outline_template, outline_name, segments)

    return str_to_jinja(template_str, template_key)



from collections import OrderedDict
import re
import os
from six import string_types


def strip_ext(path):
    return os.path.splitext(path)[0]


def basename(path, ext=False):
    basename = os.path.basename(path)
    if not ext:
        basename = os.path.splitext(basename)[0]
    return basename


def get_empty_lines(text):
    """Get number of empty lines before and after text."""
    before = len(text) - len(text.lstrip("\n"))
    after = len(text) - len(text.strip("\n")) - before
    return before, after


def wrap_latex(input, max_length=75, **kwargs):
    if len(input) > max_length:
        # remove double dollars, as they don't allow word wrap
        if len(input) > 3:
            if input[0:2] == "$$" and input[-2:] == "$$":
                input = input[1:-1]
        # change \left( and \right) to \bigg( and \bigg), as allow word wrap
        input = input.replace(r"\left(", r"\big(")
        input = input.replace(r"\right)", r"\big)")

    return input


def remove_dollars(text):
    """remove dollars from start/end of text"""
    while text.startswith("$"):
        text = text[1:]
    while text.endswith("$"):
        text = text[0:-1]
    return text


def wrap_eqn(text, cell_meta, nb_meta, out="latex"):
    """ wrap an equation in a latex equation environment

    environment obtained (if present) from cell_meta.ipub.equation.environment)
    label obtained (if present) from cell_meta.ipub.equation.label)
    also, nb_meta.ipub.enable_breqn used

    Parameters
    ----------
    text: str
    cell_meta: dict
        the cell metadata
    nb_meta: dict
        the notebook metadata

    Returns
    -------
    new_text: str

    """
    numbered = True
    try:
        environment = cell_meta["ipub"]["equation"]["environment"]
    except (KeyError, TypeError):
        environment = None
    if environment == "none":
        environment = None
    elif environment in ["equation*", "align*", "multline*", "gather*", "eqnarray*"]:
        numbered = False
    elif environment in ["equation", "align", "multline", "gather", "eqnarray"]:
        pass
    elif environment == "breqn" and out == "latex":
        if nb_meta.get("ipub", {}).get("enable_breqn", False):
            environment = "dmath*"
        else:
            environment = None
    else:
        environment = None

    try:
        label = cell_meta["ipub"]["equation"]["label"]
    except (KeyError, TypeError):
        label = None

    if environment:
        outtext = "\\begin{{{0}}}".format(environment)
        if label and numbered and out == "latex":
            outtext += "\\label{{{0}}}".format(label)
        outtext += "\n" + remove_dollars(text)
        outtext += "\n\\end{{{0}}}".format(environment)
    else:
        outtext = text

    return outtext


def get_caption(etype, cell_meta, resources):
    """return an ipypublish caption or False

    captions can either be located at cell_meta.ipub.<type>.caption,
    or at resources.caption[cell_meta.ipub.<type>.label]

    the resources version is proritised
    """
    try:
        caption = cell_meta["ipub"][etype]["caption"]
    except (KeyError, TypeError):
        caption = False

    try:
        label = cell_meta["ipub"][etype]["label"]
    except (KeyError, TypeError):
        label = False

    rcaption = False
    if label:
        try:
            rcaption = resources["caption"][label]
        except (KeyError, TypeError):
            pass

    if rcaption:
        return rcaption
    return caption


def first_para(input, **kwargs):
    r"""get only ttext before a \n (i.e. the fist paragraph)"""
    return input.split("\n")[0]


def _write_roman(num):
    roman = OrderedDict()
    roman[1000] = "M"
    roman[900] = "CM"
    roman[500] = "D"
    roman[400] = "CD"
    roman[100] = "C"
    roman[90] = "XC"
    roman[50] = "L"
    roman[40] = "XL"
    roman[10] = "X"
    roman[9] = "IX"
    roman[5] = "V"
    roman[4] = "IV"
    roman[1] = "I"

    def roman_num(num):
        for r in roman.keys():
            x, y = divmod(num, r)
            yield roman[r] * x
            num -= r * x
            if num > 0:
                roman_num(num)
            else:
                break

    return "".join([a for a in roman_num(num)])


def _repl(match):
    return _write_roman(int(match.group(0)))


def create_key(input, **kwargs):
    """create sanitized key string which only contains lowercase letters,
    (semi)colons as c, underscores as u and numbers as roman numerals
    in this way the keys with different input should mainly be unique

    >>> create_key('fig:A_10name56')
    'figcauxnamelvi'

    """
    input = re.compile(r"\d+").sub(_repl, input)
    input = input.replace(":", "c")
    input = input.replace(";", "c")
    input = input.replace("_", "u")
    return re.sub("[^a-zA-Z]+", "", str(input)).lower()


def _split_option(item, original):
    opt = item.split("=")
    if len(opt) > 2:
        raise ValueError(
            "item '{}' from '{}' contains multiple '='".format(item, original)
        )
    elif len(opt) == 1:
        return opt[0].strip(), None
    else:
        return [o.strip() for o in opt]


def dict_to_kwds(inobject, kwdstr="", overwrite=True):
    """ convert a dictionary to a string of keywords,
    or, if a list, a string of options

    append to an existing options string (without duplication)

    Parameters
    ----------
    dct : dict
    kwdstr: str
        initial keyword string
    overwrite: bool
        overwrite the option, if it already exists with a different value

    Examples
    --------
    >>> dict_to_kwds({"a":1,"c":3},'a=1,b=2')
    'a=1,b=2,c=3'
    >>> dict_to_kwds(['a', 'c'],'a,b')
    'a,b,c'

    """
    if not isinstance(kwdstr, string_types):
        raise ValueError("kwdstr '{}' not a string".format(kwdstr))

    optdict = {}
    for item in kwdstr.split(","):
        if item == "":
            continue
        ikey, ival = _split_option(item, kwdstr)
        if ikey in optdict:
            raise ValueError(
                "kwdstr '{}' contain multiple references to '{}'".format(kwdstr, ikey)
            )
        optdict[ikey] = ival

    if isinstance(inobject, (list, tuple)):
        for item in inobject:
            if item == "":
                continue
            if not isinstance(item, string_types):
                raise ValueError(
                    "option '{}' from option list is not a string: {}".format(
                        item, kwdstr
                    )
                )
            okey, oval = _split_option(item, inobject)
            if okey not in optdict or overwrite:
                optdict[okey] = oval
    else:
        for kkey in sorted(inobject.keys()):
            keystr = str(kkey)
            if keystr not in optdict or overwrite:
                optdict[kkey] = str(inobject[kkey])

    outstring1 = []
    outstring2 = []
    for skey in sorted(optdict.keys()):
        if optdict[skey] is None:
            outstring1.append(skey)
        else:
            outstring2.append("{}={}".format(skey, optdict[skey]))

    outstring = outstring1 + outstring2
    return ",".join(outstring)


def is_equation(text):
    """test if a piece of text is a latex equation, by how it is wrapped"""
    text = text.strip()

    if any(
        [
            text.startswith("\\begin{{{0}}}".format(env))
            and text.endswith("\\end{{{0}}}".format(env))
            for env in [
                "equation",
                "split",
                "equation*",
                "align",
                "align*",
                "multline",
                "multline*",
                "gather",
                "gather*",
            ]
        ]
    ):
        return True
    elif text.startswith("$") and text.endswith("$"):
        return True
    else:
        return False


if __name__ == "__main__":

    print(dict_to_kwds(["a", "c"], "e,b,d=3"))



import io
import re
import string

from six import string_types
import bibtexparser


def read_bibliography(path, raise_error=True):
    """ read a bibliography

    """
    bibdatabase = {}
    bibparser = bibtexparser.bparser.BibTexParser()
    try:
        if hasattr(path, "open"):
            with path.open(encoding="utf8") as bibtex_file:
                bibtex_data = bibtex_file.read()
        else:
            with io.open(path, encoding="utf8") as bibtex_file:
                bibtex_data = bibtex_file.read()
        bibtex_data = safe_str(bibtex_data)
        bibdatabase = bibparser.parse(bibtex_data).entries_dict
    except Exception as err:
        if raise_error:
            raise IOError("could not read bibliopath {}: {}".format(path, err))

    return bibdatabase


def process_bib_entry(
    cid, bibdatabase, bibnums, fallback_fmt="[{author_abbrev}, {year}]"
):
    """work out the best way to represent the bib entry """
    entry = bibdatabase[cid]
    if cid not in bibnums:
        bibnums[cid] = len(bibnums) + 1

    if "doi" in entry:
        return r'<a href="https://doi.org/{doi}">{text}</a>'.format(
            doi=entry["doi"], text=bibnums[cid]
        )
    elif "url" in entry:
        return r'<a href="{url}">{text}</a>'.format(url=entry["url"], text=bibnums[cid])
    elif "link" in entry:
        return r'<a href="{url}">{text}</a>'.format(
            url=entry["link"], text=bibnums[cid]
        )
    else:
        return bibnums[cid]
        # add_abbreviated_author(entry)
        # split_date(entry)
        # return DefaultFormatter().format(fallback_fmt, **entry)


def add_abbreviated_author(entry):
    # abbreviate a list of authors
    if "author" in entry:
        authors = re.split(", | and ", entry["author"])
        if len(authors) > 1:
            author_abbrev = authors[0] + " <em>et al</em>"
        else:
            author_abbrev = authors[0]
        entry["author_abbrev"] = author_abbrev


def split_date(entry):
    # split up date into year, month, day
    if "date" in entry:
        date = entry["date"].split("-")
        if len(date) == 3:
            entry["year"] = date[0]
            entry["month"] = date[1]
            entry["day"] = date[2]
        else:
            entry["year"] = date[0]


class DefaultFormatter(string.Formatter):
    def __init__(self, default=""):
        self.default = default

    def get_value(self, key, args, kwds):
        if isinstance(key, string_types):
            return kwds.get(key, self.default.format(key))
        else:
            string.Formatter.get_value(key, args, kwds)


def safe_str(obj):
    if hasattr(obj, "decode"):
        try:
            obj = obj.decode("utf-8")
        except UnicodeEncodeError:
            pass
    try:
        return str(obj)
    except UnicodeEncodeError:
        # python 2.7
        obj = re.sub(u"\u2013", "-", obj)  # en dash
        obj = re.sub(u"\u2014", "--", obj)  # em dash
        return obj.encode("ascii", "ignore").decode("ascii")
    return ""



from collections import OrderedDict
import copy
import io
import json
import re

from six import string_types
from nbconvert.utils.pandoc import get_pandoc_version
from distutils.version import LooseVersion
import panflute as pf

from panflute import Element, Doc  # noqa: F401
from types import FunctionType  # noqa: F401

from ipypublish.filters_pandoc.definitions import IPUB_META_ROUTE


def apply_filter(
    in_object,
    filter_func=None,
    out_format="panflute",
    in_format="markdown",
    strip_meta=False,
    strip_blank_lines=False,
    replace_api_version=True,
    dry_run=False,
    **kwargs
):
    # type: (list[str], FunctionType) -> str
    """convenience function to apply a panflute filter(s)
    to a string, list of string lines, pandoc AST or panflute.Doc

    Parameters
    ----------
    in_object: str or list[str] or dict
        can also be panflute.Doc
    filter_func:
        the filter function or a list of filter functions
    out_format: str
        for use by pandoc or, if 'panflute', return the panflute.Doc
    in_format="markdown": str
    strip_meta=False: bool
        strip the document metadata before final conversion
    strip_blank_lines: bool
    strip_ends: bool
        strip any blank lines or space from the start and end
    replace_api_version: bool
        for dict input only, if True,
        find the api_version of the available pandoc and
        reformat the json as appropriate
    dry_run: bool
        If True, return the Doc object, before applying the filter
    kwargs:
        to parse to filter func

    Returns
    -------
    str

    """
    if isinstance(in_object, pf.Doc):
        pass
    elif isinstance(in_object, dict):
        if not in_format == "json":
            raise AssertionError(
                "the in_format for a dict should be json, " "not {}".format(in_format)
            )
        if "meta" not in in_object:
            raise ValueError("the in_object does contain a 'meta' key")
        if "blocks" not in in_object:
            raise ValueError("the in_object does contain a 'blocks' key")
        if "pandoc-api-version" not in in_object:
            raise ValueError("the in_object does contain a 'pandoc-api-version' key")
        if replace_api_version:
            # run pandoc on a null object, to get the correct api version
            null_raw = pf.run_pandoc("", args=["-t", "json"])
            null_stream = io.StringIO(null_raw)
            api_version = pf.load(null_stream).api_version

            # see panflute.load, w.r.t to legacy version
            if api_version is None:
                in_object = [{"unMeta": in_object["meta"]}, in_object["blocks"]]
            else:
                ans = OrderedDict()
                ans["pandoc-api-version"] = api_version
                ans["meta"] = in_object["meta"]
                ans["blocks"] = in_object["blocks"]
                in_object = ans
        in_str = json.dumps(in_object)
    elif isinstance(in_object, (list, tuple)):
        in_str = "\n".join(in_object)
    elif isinstance(in_object, string_types):
        in_str = in_object
    else:
        raise TypeError("object not accepted: {}".format(in_object))

    if not isinstance(in_object, pf.Doc):
        doc = pf.convert_text(in_str, input_format=in_format, standalone=True)
        # f = io.StringIO(in_json)
        # doc = pf.load(f)
    else:
        doc = in_object

    doc.format = out_format

    if dry_run:
        return doc

    if not isinstance(filter_func, (list, tuple, set)):
        filter_func = [filter_func]

    out_doc = doc
    for func in filter_func:
        out_doc = func(out_doc, **kwargs)  # type: Doc

    # post-process Doc
    if strip_meta:
        out_doc.metadata = {}
    if out_format == "panflute":
        return out_doc

    # create out str
    # with io.StringIO() as f:
    #     pf.dump(doc, f)
    #     jsonstr = f.getvalue()
    # jsonstr = json.dumps(out_doc.to_json()
    out_str = pf.convert_text(
        out_doc, input_format="panflute", output_format=out_format
    )

    # post-process final str
    if strip_blank_lines:
        out_str = out_str.replace("\n\n", "\n")

    return out_str


def compare_version(target, comparison):
    """Set docstring here.

    Parameters
    ----------
    target: str
        target version of pandoc
    comparison: str
        one of '>', '<', '<=', '>=', '=='

    Returns
    -------
    bool

    """
    # TODO this only works if you are
    # converting json in the same environment
    # from pandocxnos import init as get_pandoc_version
    version = LooseVersion(get_pandoc_version())
    required = LooseVersion(target)
    if comparison == ">=":
        return version >= required
    elif comparison == "<=":
        return version <= required
    elif comparison == ">":
        return version > required
    elif comparison == "<":
        return version < required
    elif comparison == "==":
        return version == required
    else:
        raise ValueError("comparison not recognised: {}".format(comparison))


def strip_quotes(string):
    # type: (str) -> str
    if string.startswith("'") and string.endswith("'"):
        string = string[1:-1]
    if string.startswith('"') and string.endswith('"'):
        string = string[1:-1]
    return string


def find_attributes(
    element, allow_space=True, search_left=False, include_element=False
):
    """find an attribute 'container' for an element,
    of the form <element><space>{#id .class1 .class2 a=1 b="a string"}
    and extract its content

    Parameters
    ----------
    element:
        the element to find attributes for
    allow_space=True: bool
        whether to allow space between the element and attribute container
    search_left=False: bool
        search to the left of the element, rather than the right
    include_element=False: bool
        whether to include the element in the search


    Returns
    -------
    dict or None:
        {"classes": list[str], "attributes": dict[str],
         "id": str, "elements": list[Element]}, where elements is
         the elements containing the attributes (including space)

    """
    if search_left:
        return _search_attribute_left(element, include_element, allow_space)
    else:
        return _search_attribute_right(element, include_element, allow_space)


def _search_attribute_right(element, include_element, allow_space):
    if (not element.next) and not include_element:
        return None

    if include_element:
        adjacent = element
    else:
        adjacent = element.next

    attr_elements = []
    found_start = False
    found_end = False
    while adjacent:
        if isinstance(adjacent, pf.Space) and allow_space:
            attr_elements.append(adjacent)
            adjacent = adjacent.next
            continue
        elif (
            isinstance(adjacent, pf.Str)
            #   and adjacent.text.startswith("{")
            #   and adjacent.text.endswith("}")):
            and re.search(r"^\{[^}]*\}", adjacent.text)
        ):
            # TODO this won't handle } in strings, e.g. {a="} "}
            found_start = True
            found_end = True
            attr_elements.append(adjacent)
            break
        elif (
            isinstance(adjacent, pf.Str)
            #   and adjacent.text.startswith("{")):
            and re.search(r"^[^\}]*\{", adjacent.text)
        ):
            found_start = True
            found_end = False
            attr_elements.append(adjacent)
            break
        break
        # adjacent = adjacent.next

    if found_start and not found_end:
        adjacent = adjacent.next
        while adjacent:
            if (
                isinstance(adjacent, pf.Str)
                # and adjacent.text.endswith("}")):
                and re.search(r"^[^\{]*\}", adjacent.text)
            ):
                # TODO this won't handle } in strings, e.g. {a="} "}
                found_end = True
                attr_elements.append(adjacent)
                break
            else:
                attr_elements.append(adjacent)
            adjacent = adjacent.next

    if not (found_start and found_end):
        return None

    attribute_str = pf.stringify(pf.Para(*attr_elements)).replace("\n", " ").strip()

    # split into the label and the rest
    match = re.match(r"^\{(#[^\s]+|)([^\}]*)\}", attribute_str)
    if not match:
        raise ValueError(attribute_str)
    classes, attributes = process_attributes(match.group(2))

    new_str = attribute_str[len(match.group(0)) :]

    return {
        "id": match.group(1)[1:],
        "classes": classes,
        "attributes": attributes,
        "elements": attr_elements,
        "append": pf.Str(new_str) if new_str else None,
    }


def _search_attribute_left(element, include_element, allow_space):
    if (not element.prev) and not include_element:
        return None

    if include_element:
        adjacent = element
    else:
        adjacent = element.prev

    attr_elements = []
    found_start = False
    found_end = False
    while adjacent:
        if isinstance(adjacent, pf.Space) and allow_space:
            attr_elements.append(adjacent)
            adjacent = adjacent.prev
            continue
        elif (
            isinstance(adjacent, pf.Str)
            and adjacent.text.endswith("}")
            and adjacent.text.startswith("{")
        ):
            # TODO this won't handle } in strings, e.g. {a="} "}
            # TODO this won't handle characters after } e.g. {a=1})
            found_start = True
            found_end = True
            attr_elements.append(adjacent)
            break
        elif isinstance(adjacent, pf.Str) and adjacent.text.endswith("}"):
            found_start = False
            found_end = True
            attr_elements.append(adjacent)
            break
        break
        # adjacent = adjacent.prev

    if found_end and not found_start:
        adjacent = adjacent.prev
        while adjacent:
            if isinstance(adjacent, pf.Str) and adjacent.text.startswith("{"):
                # TODO this won't handle { in strings, e.g. {a="{ "}
                # TODO this won't handle characters before { e.g. ({a=1}
                found_start = True
                attr_elements.append(adjacent)
                break
            else:
                attr_elements.append(adjacent)
            adjacent = adjacent.prev

    if not (found_start and found_end):
        return None

    attr_elements = list(reversed(attr_elements))

    attribute_str = pf.stringify(pf.Para(*attr_elements)).replace("\n", " ").strip()

    # split into the label and the rest
    match = re.match("^\\{(#[^\\s]+|)([^\\}]*)\\}$", attribute_str)
    if not match:
        raise ValueError(attribute_str)
    classes, attributes = process_attributes(match.group(2))

    return {
        "id": match.group(1)[1:],
        "classes": classes,
        "attributes": attributes,
        "elements": attr_elements,
        "append": None,
    }


def process_attributes(attr_string):
    """process a string of classes and attributes,
    e.g. '.class-name .other a=1 b="some text"' will be returned as:
    ["class-name", "other"], {"a": 1, "b": "some text"}

    Returns:
    list: classes
    dict: attributes
    """
    # find classes, denoted by .class-name
    classes = [c[1][1:] for c in re.findall("(^|\\s)(\\.[\\-\\_a-zA-Z]+)", attr_string)]
    # find attributes, denoted by a=b, respecting quotes
    attr = {
        c[1]: strip_quotes(c[2])
        for c in re.findall(
            "(^|\\s)([\\-\\_a-zA-Z]+)\\s*=\\s*(\\\".+\\\"|\\'.+\\'|[^\\s\\\"\\']+)",  # noqa: E501
            attr_string,
        )
    }
    # TODO this generally works, but should be stricter against any weird
    # fringe cases

    # TODO add tests
    return classes, attr


def convert_attributes(attr):
    """attempt to convert values to python types, e.g. float, list, dict"""
    attr = copy.deepcopy(attr)
    for key in list(attr.keys()):
        try:
            new_value = json.loads(attr[key])
            attr[key] = new_value
        except Exception:
            pass
    return attr


def convert_units(string, out_units):
    match = re.compile("^\\s*([0-9]+\\.?[0-9]*)([a-z\\%]*)\\s*$").match(str(string))
    if match is None:
        raise ValueError("string could not be resolved as a value: {}".format(string))
    value = float(match.group(1))
    in_units = match.group(2)
    in_units = "fraction" if not in_units else in_units

    if in_units == out_units:
        return value

    convert = {
        ("%", "fraction"): lambda x: x / 100.0,
        ("fraction", "%"): lambda x: x * 100.0,
    }.get((in_units, out_units), None)

    if convert is None:
        raise ValueError(
            "could not find a conversion for "
            "{0} to {1}: {2}".format(in_units, out_units, string)
        )

    return convert(value)


def get_option(locations, keypath, default=None, delimiter=".", error_on_missing=False):
    """ fetch an option variable from a hierarchy of preferred locations

    The value returned will be from the first available location or the default

    Parameters
    ----------
    locations: list[dict]
        a list of mappings to search in
    keypath: list[str] or str
        a key path to search in, if str, then split by delimiter
    default=None: object
        a default value to return
    delimiter: str
        if a str then the keypath is expected to be a str
    error_on_missing: bool
        raise KeyError if not found in any of the options

    Examples
    --------

    >>> a = {"m": 1}
    >>> b = {"x": {"y": 2}}
    >>> c = {"x": {"y": 3}}
    >>> get_option([a, b, c], keypath=("x", "y"))
    2
    >>> get_option([a, c, b], keypath=("x", "y"))
    3
    >>> get_option([a, c, b], keypath="x.y")
    3
    >>> get_option([a, c, b], keypath="l", default=4)
    4

    """
    if isinstance(keypath, string_types):
        keypath = keypath.split(delimiter)

    found_var = False
    variable = None

    for opt in locations:
        final_opt = opt
        found_key = True
        for key in keypath:
            try:
                final_opt = final_opt[key]
            except (KeyError, TypeError):
                found_key = False
                break
        if found_key:
            found_var = True
            variable = final_opt
            break

    if found_var:
        return variable
    elif error_on_missing:
        raise ValueError("could not retrieve the option keypath: {}".format(keypath))

    return default


def create_ipub_meta(options):
    meta = {}
    submeta = meta
    for key in IPUB_META_ROUTE.split(".")[:-1]:
        submeta[key] = {}
        submeta = submeta[key]
    submeta[IPUB_META_ROUTE.split(".")[-1]] = options
    return meta


def get_panflute_containers(element):
    """return list of all possible container classes for an element"""
    panflute_inline_containers = (
        pf.Cite,
        pf.Emph,
        pf.Header,
        pf.Image,
        pf.LineItem,
        pf.Link,
        pf.Para,
        pf.Plain,
        pf.Quoted,
        pf.SmallCaps,
        pf.Span,
        pf.Strikeout,
        pf.Strong,
        pf.Subscript,
        pf.Superscript,
    )

    panflute_block_containers = (
        pf.BlockQuote,
        pf.Definition,
        pf.Div,
        pf.Doc,
        pf.ListItem,
        pf.Note,
        pf.TableCell,
    )

    if issubclass(element, pf.Inline):
        return panflute_inline_containers

    elif issubclass(element, pf.Block):
        return panflute_block_containers

    raise TypeError("not Inline or Block: {}".format(element))


def get_pf_content_attr(container, target):

    panflute_inline_containers = [
        pf.Cite,
        pf.Emph,
        pf.Header,
        pf.Image,
        pf.LineItem,
        pf.Link,
        pf.Para,
        pf.Plain,
        pf.Quoted,
        pf.SmallCaps,
        pf.Span,
        pf.Strikeout,
        pf.Strong,
        pf.Subscript,
        pf.Superscript,
        pf.Table,
        pf.DefinitionItem,
    ]

    panflute_block_containers = (
        pf.BlockQuote,
        pf.Definition,
        pf.Div,
        pf.Doc,
        pf.ListItem,
        pf.Note,
        pf.TableCell,
    )

    if issubclass(target, pf.Cite):
        # we assume a Cite can't contain another Cite
        if not isinstance(container, tuple(panflute_inline_containers[1:])):
            return False

    if issubclass(target, pf.Inline):
        if isinstance(container, tuple(panflute_inline_containers)):
            if isinstance(container, pf.Table):
                return "caption"
            elif isinstance(container, pf.DefinitionItem):
                return "term"
            else:
                return "content"
        else:
            return False

    if issubclass(target, pf.Block):
        if isinstance(container, tuple(panflute_block_containers)):
            return "content"
        else:
            return False

    raise TypeError("target not Inline or Block: {}".format(target))



import os
import pytest
from ipypublish.utils import pathlib  # noqa: F401
from ipypublish.frontend import nbpresent
from ipypublish.frontend import nbpublish


@pytest.mark.ipynb("basic_nb")
def test_nbpresent_bad_exporter(ipynb_app):
    # type: (str, pathlib.Path) -> None
    assert 1 == nbpresent.run(
        [
            str(ipynb_app.input_file),
            "-f",
            "non-existent",
            "--outpath",
            str(ipynb_app.converted_path),
            "--dry-run",
            "--log-level",
            "debug",
        ]
    )


@pytest.mark.ipynb("basic_nb")
def test_nbpublish_bad_exporter(ipynb_app):
    # type: (str, pathlib.Path) -> None
    assert 1 == nbpublish.run(
        [
            str(ipynb_app.input_file),
            "-f",
            "non-existent",
            "--outpath",
            str(ipynb_app.converted_path),
            "--dry-run",
            "--log-level",
            "debug",
        ]
    )


@pytest.mark.ipynb("basic_nb")
def test_nbpresent_dry_run(ipynb_app):
    # type: (str, pathlib.Path) -> None
    assert 0 == nbpresent.run(
        [
            str(ipynb_app.input_file),
            "--outpath",
            str(ipynb_app.converted_path),
            "--dry-run",
            "--log-level",
            "debug",
            "-pt",
        ]
    )


@pytest.mark.ipynb("basic_nb")
def test_nbpublish_dry_run(ipynb_app):
    # type: (str, pathlib.Path) -> None
    assert 0 == nbpublish.run(
        [
            str(ipynb_app.input_file),
            "--outpath",
            str(ipynb_app.converted_path),
            "--dry-run",
            "--log-level",
            "debug",
            "-pt",
        ]
    )


@pytest.mark.ipynb("basic_nb")
def test_nbpublish_dry_run_with_external_plugin(ipynb_app, external_export_plugin):
    # type: (str, pathlib.Path) -> None
    assert 0 == nbpublish.run(
        [
            str(ipynb_app.input_file),
            "--outformat",
            str(external_export_plugin),
            "--outpath",
            str(ipynb_app.converted_path),
            "--dry-run",
            "--log-level",
            "debug",
            "-pt",
        ]
    )


@pytest.mark.ipynb("basic_nb")
def test_nbpublish_dry_run_with_external_plugin_key(ipynb_app, external_export_plugin):
    # type: (str, pathlib.Path, pathlib.Path) -> None
    assert 0 == nbpublish.run(
        [
            str(ipynb_app.input_file),
            "--export-paths",
            str(external_export_plugin.parent),
            "--outformat",
            os.path.splitext(str(external_export_plugin.name))[0],
            "--outpath",
            str(ipynb_app.converted_path),
            "--dry-run",
            "--log-level",
            "debug",
            "-pt",
        ]
    )


def test_nbpresent_list_exports():
    with pytest.raises(SystemExit) as out:
        nbpresent.run(["--list-exporters"])
        assert out.type == SystemExit
        assert out.value.code == 0


def test_nbpublish_list_exports():
    with pytest.raises(SystemExit) as out:
        nbpublish.run(["--list-exporters"])
        assert out.type == SystemExit
        assert out.value.code == 0


@pytest.mark.ipynb("basic_nb")
def test_nbpublish_write(ipynb_app):
    # type: (str, pathlib.Path) -> None
    assert 0 == nbpublish.run(
        [
            str(ipynb_app.input_file),
            "--outformat",
            "latex_ipypublish_main",
            "--outpath",
            str(ipynb_app.converted_path),
            "-pt",
        ]
    )
    assert ipynb_app.converted_path.joinpath(
        ipynb_app.input_file.name.replace(".ipynb", ".tex")
    ).exists()


@pytest.mark.requires_latexmk
@pytest.mark.ipynb("basic_nb")
def test_nbpublish_to_pdf(ipynb_app):
    # type: (str, pathlib.Path) -> None
    assert 0 == nbpublish.run(
        [
            str(ipynb_app.input_file),
            "--outformat",
            "latex_ipypublish_main",
            "--outpath",
            str(ipynb_app.converted_path),
            "--create-pdf",
            "-pt",
        ]
    )
    assert ipynb_app.converted_path.joinpath(
        ipynb_app.input_file.name.replace(".ipynb", ".pdf")
    ).exists()



# -*- coding: utf-8 -*-
"""
    test_sphinx
    ~~~~~~~~~~~

    General Sphinx test and check output.
"""
import pytest
import sphinx

from ipypublish.sphinx.tests import get_test_source_dir

from ipypublish.tests.utils import HTML2JSONParser


@pytest.mark.sphinx(buildername="html", srcdir=get_test_source_dir("notebook"))
def test_basic(app, status, warning, get_sphinx_app_output, data_regression):

    app.build()

    assert "build succeeded" in status.getvalue()  # Build succeeded
    warnings = warning.getvalue().strip()
    assert warnings == ""

    output = get_sphinx_app_output(app, buildername="html")

    parser = HTML2JSONParser()
    parser.feed(output)
    if sphinx.version_info >= (2,):
        data_regression.check(parser.parsed, basename="test_basic_v2")
    else:
        data_regression.check(parser.parsed, basename="test_basic_v1")


@pytest.mark.sphinx(
    buildername="html", srcdir=get_test_source_dir("notebook_cell_decor")
)
def test_cell_decoration(app, status, warning, get_sphinx_app_output, data_regression):
    """ test a notebook with prompts and toggle buttons"""
    app.build()

    assert "build succeeded" in status.getvalue()  # Build succeeded
    warnings = warning.getvalue().strip()
    assert warnings == ""

    output = get_sphinx_app_output(app, buildername="html")

    parser = HTML2JSONParser()
    parser.feed(output)
    if sphinx.version_info >= (2,):
        data_regression.check(parser.parsed, basename="test_cell_decoration_v2")
    else:
        data_regression.check(parser.parsed, basename="test_cell_decoration_v1")


@pytest.mark.sphinx(
    buildername="html", srcdir=get_test_source_dir("notebook_ipywidget")
)
def test_ipywidget(app, status, warning, get_sphinx_app_output, data_regression):
    """ test which contains an ipywidgets and the widget state has been saved."""
    app.build()

    assert "build succeeded" in status.getvalue()  # Build succeeded
    warnings = warning.getvalue().strip()
    assert warnings == ""

    output = get_sphinx_app_output(app, buildername="html")

    parser = HTML2JSONParser()
    parser.feed(output)
    data_regression.check(parser.parsed, basename="test_ipywidget")



from ipypublish import __version__
from ipypublish.sphinx.utils import import_sphinx

from ipypublish.sphinx.gls import processing as bibproc
from ipypublish.sphinx.gls.directives import BibGlossaryDirective
from ipypublish.sphinx.gls.roles import (
    GLSRole,
    GLSCapitalRole,
    GLSPluralRole,
    GLSPluralCapitalRole,
)
from ipypublish.sphinx.gls.nodes import BibGlossaryNode
from ipypublish.sphinx.gls.transforms import (
    BibGlossaryTransform,
    OverrideCitationReferences,
    HandleMissingCitesTransform,
)

try:
    from sphinx.application import Sphinx  # noqa: F401
except ImportError:
    pass


def setup(app):
    # type: (Sphinx) -> dict
    """Initialize Sphinx extension.

    """
    # delayed import of sphinx
    sphinx = import_sphinx()

    try:
        transforms = app.registry.get_transforms()
    except AttributeError:  # Sphinx < 1.7
        from sphinx.io import SphinxStandaloneReader

        transforms = SphinxStandaloneReader.transforms

    def add_transform(transform, post=False):
        if transform not in transforms:
            if post:
                app.add_post_transform(transform)
            else:
                app.add_transform(transform)

    # BibGlossary functionality
    app.connect("builder-inited", bibproc.init_bibgloss_cache)
    app.connect("doctree-resolved", bibproc.process_citations)
    app.connect("doctree-resolved", bibproc.process_citation_references)
    app.connect("env-purge-doc", bibproc.purge_bibgloss_cache)
    # app.connect("env-updated", bibproc.check_duplicate_labels)
    app.add_config_value("bibgloss_convert_latex", True, rebuild="html")
    app.add_config_value("bibgloss_default_style", "list", rebuild="html")

    app.add_directive("bibglossary", BibGlossaryDirective)
    # Note: because docutils.parsers.rst.roles.role(role_name)
    #       applies role_name.lower(), you can't have unique gls and Gls roles
    app.add_role("gls", GLSRole())
    app.add_role("glsc", GLSCapitalRole())
    app.add_role("glspl", GLSPluralRole())
    app.add_role("glscpl", GLSPluralCapitalRole())
    app.add_node(BibGlossaryNode, override=True)

    add_transform(BibGlossaryTransform)
    # these patches have been fixed (by me!) upstream
    if sphinx.version_info < (2,):
        add_transform(OverrideCitationReferences)
        add_transform(HandleMissingCitesTransform, post=True)

    # Parallel read is not safe at the moment: in the current design,
    # the document that contains references must be read last for all
    # references to be resolved.
    parallel_read_safe = False

    return {
        "version": __version__,
        "parallel_read_safe": parallel_read_safe,
        "parallel_write_safe": True,
        "env_version": 1,
    }



import string


class DefaultFormatter(string.Formatter):
    def get_value(self, key, args, kwargs):
        # Handle a key not found
        try:
            val = super(DefaultFormatter, self).get_value(key, args, kwargs)
            # Python 3, 'super().get_field(field_name, args, kwargs)' works
        except (IndexError, KeyError):
            if str(key).endswith("_pre"):
                val = "@ipubreplace{above}{" + str(key) + "}"
            else:
                val = "@ipubreplace{below}{" + str(key) + "}"
        return val


def convert_format_str(template):

    temp_str = "\n".join(template)
    fmt = DefaultFormatter()
    outstr = fmt.format(temp_str)

    return outstr


if __name__ == "__main__":

    template = [
        "{{%- extends 'null.tpl' -%}}",
        "{{% block header %}}",
        "{{{{ nb.metadata | meta2yaml('#~~ ') }}}}",
        "{{% endblock header %}}",
        "{{% block codecell %}}",
        "#%%",
        "{{{{ super() }}}}",
        "{{% endblock codecell %}}",
        "{{% block in_prompt %}}{{% endblock in_prompt %}}",
        "{{% block input %}}{{{{ cell.metadata | meta2yaml('#~~ ') }}}}",  # noqa: E501
        "{{{{ cell.source | ipython2python }}}}",
        "{{% endblock input %}}",
        "{{% block markdowncell scoped %}}#%% [markdown]",
        "{{{{ cell.metadata | meta2yaml('#~~ ') }}}}",
        "{{{{ cell.source | comment_lines }}}}",
        "{{% endblock markdowncell %}}",
    ]

    print(convert_format_str(template))



import sys
import pytest

if sys.version_info >= (3, 0):
    from ipypublish.bib2glossary.parse_tex import parse_tex


@pytest.mark.skipif(
    sys.version_info < (3, 0),
    reason="SyntaxError on import of texsoup/data.py line 135",
)
def test_parse_acronyms():

    text_str = """
    \\newacronym{otherkey}{OTHER}{Abbreviation of other}
    \\newacronym{thekey}{ABRV}{Abbreviation}
    """
    gterms, acronyms = parse_tex(text_str=text_str)
    assert gterms == {}
    assert acronyms == {
        "otherkey": {"abbreviation": "OTHER", "longname": "Abbreviation of other"},
        "thekey": {"abbreviation": "ABRV", "longname": "Abbreviation"},
    }


@pytest.mark.skipif(
    sys.version_info < (3, 0),
    reason="SyntaxError on import of texsoup/data.py line 135",
)
def test_parse_acronyms_with_options():

    text_str = """
    \\newacronym[description={a description}]{otherkey}{OTHER}{Abbreviation of other}
    \\newacronym[plural={ABRVs},longplural={Abbreviations}]{thekey}{ABRV}{Abbreviation}
    """  # noqa: E501
    gterms, acronyms = parse_tex(text_str=text_str)
    assert gterms == {}
    assert acronyms == {
        "otherkey": {
            "abbreviation": "OTHER",
            "longname": "Abbreviation of other",
            "description": "a description",
        },
        "thekey": {
            "abbreviation": "ABRV",
            "longname": "Abbreviation",
            "longplural": "Abbreviations",
            "plural": "ABRVs",
        },
    }


@pytest.mark.skipif(
    sys.version_info < (3, 0),
    reason="SyntaxError on import of texsoup/data.py line 135",
)
def test_parse_gterms():

    text_str = """
    \\newglossaryentry{otherkey}{
        name={other name},
        description={the description of other}
    }
    \\newglossaryentry{thekey}{
        name={name},
        description={the description},
        type={symbols}
    }
    """
    gterms, acronyms = parse_tex(text_str=text_str)
    assert acronyms == {}
    assert gterms == {
        "otherkey": {"description": "the description of other", "name": "other name"},
        "thekey": {"description": "the description", "name": "name", "type": "symbols"},
    }


@pytest.mark.skipif(
    sys.version_info < (3, 0),
    reason="SyntaxError on import of texsoup/data.py line 135",
)
def test_parse_mixed():
    text_str = """
    \\newacronym{otherkey}{OTHER}{Abbreviation of other}
    \\newglossaryentry{thekey}{
        name={name},
        description={the description},
        type={symbols}
    }
    """
    gterms, acronyms = parse_tex(text_str=text_str)
    assert acronyms == {
        "otherkey": {"abbreviation": "OTHER", "longname": "Abbreviation of other"}
    }
    assert gterms == {
        "thekey": {"description": "the description", "name": "name", "type": "symbols"}
    }


@pytest.mark.skipif(
    sys.version_info < (3, 0),
    reason="SyntaxError on import of texsoup/data.py line 135",
)
def test_duplicate_key():
    text_str = """
    \\newacronym{thekey}{OTHER}{Abbreviation of other}
    \\newglossaryentry{thekey}{
        name={name},
        description={the description},
        type={symbols}
    }
    """
    with pytest.raises(KeyError):
        parse_tex(text_str=text_str)


@pytest.mark.skipif(
    sys.version_info < (3, 0),
    reason="SyntaxError on import of texsoup/data.py line 135",
)
def test_acronym_ioerror():
    text_str = """
    \\newacronym{thekey}{Abbreviation of other}
    """
    with pytest.raises(IOError):
        parse_tex(text_str=text_str)


@pytest.mark.skipif(
    sys.version_info < (3, 0),
    reason="SyntaxError on import of texsoup/data.py line 135",
)
def test_gterm_ioerror():
    text_str = """
    \\newglossaryentry{}
    """
    with pytest.raises(IOError):
        parse_tex(text_str=text_str)


@pytest.mark.skipif(
    sys.version_info < (3, 0),
    reason="SyntaxError on import of texsoup/data.py line 135",
)
def test_ioerror_skip():
    text_str = """
    \\newacronym{thekey}{Abbreviation of other}
    \\newacronym{thekey2}{ABBR}{Abbreviation of other}
    """
    gterms, acronyms = parse_tex(text_str=text_str, skip_ioerrors=True)
    assert gterms == {}
    assert acronyms == {
        "thekey2": {"abbreviation": "ABBR", "longname": "Abbreviation of other"}
    }



#!/usr/bin/env python
"""
create template

philosophy is only turn stuff on when we want

http://nbconvert.readthedocs.io/en/latest/customizing.html#Template-structure
http://nbconvert.readthedocs.io/en/latest/api/exporters.html#nbconvert.exporters.TemplateExporter

"""
from typing import List, Tuple, Union  # noqa: F401
import re
import io
import logging
import jsonschema

from six import string_types

# from ipypublish import __version__
from ipypublish import schema
from ipypublish.utils import handle_error, read_file_from_directory, get_module_path

logger = logging.getLogger("template")

_SEGMENT_SCHEMA_FILE = "segment.schema.json"
_SEGMENT_SCHEMA = None


def multireplace(string, replacements):
    """
    Given a string and a replacement map, it returns the replaced string.

    From https://gist.github.com/bgusach/a967e0587d6e01e889fd1d776c5f3729

    :param str string: string to execute replacements on
    :param dict replacements: replacement dictionary {find value: replacement}
    :rtype: str

    """
    if not replacements:
        return string

    # Place longer ones first to keep shorter substrings from matching,
    # where the longer ones should take place
    # For instance given the replacements {'ab': 'AB', 'abc': 'ABC'}
    # against the string 'hey abc', it should produce
    # 'hey ABC' and not 'hey ABc'
    substrs = sorted(replacements, key=len, reverse=True)

    # Create a big OR regex that matches any of the substrings to replace
    regexp = re.compile("|".join(map(re.escape, substrs)))

    # For each match, look up the new string in the replacements
    return regexp.sub(lambda match: replacements[match.group(0)], string)


def _output_to_file(content, outpath):
    if outpath is not None:
        with io.open(outpath, "w", encoding="utf8") as f:  # TODO use pathlib
            f.write(content)
        return


def create_template(outline_template, outline_name, segment_datas, outpath=None):
    # type: (dict, Tuple[dict]) -> str
    """ build a latex jinja template from;

    - a jinja(2) template outline,
      which may contain segment placeholders,
    - and json segment files adhering to the segment.schema.json schema

    if a segment contains the key "overwrite",
    then its value should be a list of keys,
    such that these key values overwrite any entries before

    Parameters
    ----------
    outline_template: str
    segment_datas: tuple or dict
    outpath:  None or str
        if not None, output to path

    """
    # get the placeholders @ipubreplace{above|below}{name}
    regex = re.compile("\\@ipubreplace\\{([^\\}]+)\\}\\{([^\\}]+)\\}", re.MULTILINE)
    placeholder_tuple = regex.findall(outline_template)

    if not placeholder_tuple:
        if segment_datas:
            handle_error(
                "the segment data is provided, "
                + "but the outline template contains no placeholders",
                KeyError,
                logger,
            )

        if outpath:
            _output_to_file(outline_template, outpath)
        return outline_template

    placeholders = {name: append for append, name in placeholder_tuple}
    # TODO validate that placeholders to not exist multiple times,
    # with above and below

    replacements = {key: "" for key in placeholders.keys()}
    docstrings = ["outline: {}".format(outline_name)]

    if segment_datas:
        docstrings.append("with segments:")
        global _SEGMENT_SCHEMA
        if _SEGMENT_SCHEMA is None:
            # lazy segment schema once
            _SEGMENT_SCHEMA = read_file_from_directory(
                get_module_path(schema),
                _SEGMENT_SCHEMA_FILE,
                "segment configuration schema",
                logger,
                interp_ext=True,
            )

    for seg_num, segment_data in enumerate(segment_datas):

        # validate segment
        try:
            jsonschema.validate(segment_data, _SEGMENT_SCHEMA)
        except jsonschema.ValidationError as err:
            handle_error(
                "validation of template segment {} failed: {}".format(
                    seg_num, err.message
                ),
                jsonschema.ValidationError,
                logger=logger,
            )

        # get description of segment
        docstrings.append(
            "- {0}: {1}".format(segment_data["identifier"], segment_data["description"])
        )

        # find what key to overwrite
        overwrite = segment_data.get("overwrite", [])
        logger.debug("overwrite keys: {}".format(overwrite))

        for key, segtext in segment_data.get("segments").items():

            if key not in placeholders:
                handle_error(
                    "the segment key '{}' ".format(key)
                    + "is not contained in the outline template",
                    KeyError,
                    logger,
                )

            if not isinstance(segtext, string_types):
                segtext = "\n".join(segtext)
            if key in overwrite:
                replacements[key] = segtext
            elif placeholders[key] == "above":
                replacements[key] = segtext + "\n" + replacements[key]
            elif placeholders[key] == "below":
                replacements[key] = replacements[key] + "\n" + segtext
            else:
                handle_error(
                    (
                        "the placeholder @ipubreplace{{{0}}}{{{1}}} ".format(
                            key, placeholders[key]
                        )
                        + "should specify 'above' or 'below' appending"
                    ),
                    jsonschema.ValidationError,
                    logger=logger,
                )

    if "meta_docstring" in placeholders:
        docstring = "\n".join([s for s in docstrings if s]).replace("'", '"')
        replacements["meta_docstring"] = docstring
    if "ipypub_version" in placeholders:
        # TODO add option to include ipypub version in output file
        # not included by default,
        # since tests need to be changed to ignore version number
        replacements["ipypub_version"] = ""  # str(__version__)

    prefix = "@ipubreplace{"
    replace_dict = {
        prefix + append + "}{" + name + "}": replacements.get(name, "")
        for append, name in placeholder_tuple
    }
    outline = multireplace(outline_template, replace_dict)

    if outpath:
        _output_to_file(outline, outpath)

    return outline



from ipypublish.filters_pandoc.utils import apply_filter
from ipypublish.filters_pandoc import prepare_cites, format_cite_elements


def test_multiple_references_rst():
    """
    """
    in_string = ["multiple references +[@fig:id; @tbl:id; @eq:id1]"]

    out_string = apply_filter(
        in_string, [prepare_cites.main, format_cite_elements.main], "rst"
    )

    print(out_string)
    assert out_string == "\n".join(
        ["multiple references :ref:`fig:id`, :ref:`tbl:id` and :ref:`eq:id1`"]
    )


def test_multiple_references_latex():
    """
    """
    in_string = ["multiple references +[@fig:id; @tbl:id; @eq:id1]"]

    out_string = apply_filter(
        in_string, [prepare_cites.main, format_cite_elements.main], "latex"
    )

    print(out_string)
    assert out_string == "\n".join(["multiple references \\cref{fig:id,tbl:id,eq:id1}"])


def test_reference_prefixes_latex():
    """
    """
    in_string = ["(?@key1 &@key2 =@key3)"]

    out_string = apply_filter(
        in_string, [prepare_cites.main, format_cite_elements.main], "latex"
    )

    print(out_string)
    assert out_string == "\n".join(["(\\Cref{key1} \\gls{key2} \\eqref{key3})"])


def test_reference_prefixes_rst():
    """
    """
    in_string = ["(?@key1 &@key2 %@key3 =@key4)"]

    out_string = apply_filter(
        in_string, [prepare_cites.main, format_cite_elements.main], "rst"
    )

    print(out_string)
    assert out_string == "\n".join(
        ["(:ref:`key1` :gls:`key2` :glsc:`key3` :eq:`key4`)"]
    )



ETYPE_GLOSS = "glsterm"
ETYPE_ACRONYM = "glsacronym"
ETYPE_SYMBOL = "glssymbol"

NEWGLOSS_FIELDS = ("name", "description", "plural", "symbol", "text", "sort")

NEWACRONYM_FIELDS = ("description", "plural", "longplural", "firstplural")


# TODO allow mapping
# DEFAULT_GLOSS_P2F = (("name", "name"),
#                      ("description", "description"),
#                      ("plural", "plural"),
#                      ("symbol", "symbol"),
#                      ("text", "text"),
#                      ("sort", "sort"))


# DEFAULT_ACRONYM_P2F = (("abbreviation", "abbreviation"),
#                        ("longname", "longname"),
#                        ("description", "description"),
#                        ("plural", "plural"),
#                        ("longplural", "longplural"),
#                        ("firstplural", "firstplural"))



import sys
from traitlets import Unicode, Enum

from ipypublish.postprocessors.base import IPyPostProcessor


class WriteStream(IPyPostProcessor):
    """ write the stream to the terminal
    """

    @property
    def allowed_mimetypes(self):
        return (
            "text/latex",
            "text/restructuredtext",
            "text/html",
            "text/x-python",
            "application/json",
            "text/markdown",
        )

    @property
    def requires_path(self):
        return True

    @property
    def logger_name(self):
        return "write-text-file"

    encoding = Unicode(
        default_value="utf8", help="the encoding of the output file"
    ).tag(config=True)

    pipe = Enum(
        ["stdout", "stderr", "stdin"],
        default_value="stdout",
        help="where to write the output to",
    ).tag(config=True)

    def run_postprocess(self, stream, mimetype, filepath, resources):

        self.logger.info("writing stream to {}".format(self.pipe))
        io_type = {"stdout": sys.stdout, "stdin": sys.stdin, "stderr": sys.stderr}.get(
            self.pipe
        )
        io_type.write(stream)

        return stream, filepath, resources



""" a panflute filter to format elements types
that may contain reference labels, i.e. Math, Image and Table.

The :py:mod:`ipypublish.filters_pandoc.prepare_labels` filter should be run
first to access the functionality below:

If the parent of the element is a Span (or Div for Table), with a class
labelled-Math/labelled-Image/labelled-Table,
then the label of the element will be span.identifier, and
the attributes and classes from this Span will be used to inform the format.

Additionally, for HTML, if a '$$references' key is available in the metadata,
this will be used to add a suffix to the element captions,
with the number of the element.

Finally, if main() is called with strip_spans = True (the default),
The Span/Div elements with classes labelled-Math/labelled-Image/labelled-Table
will be stripped from the document

"""
# TODO format headers with section labels
# (see ipysphinx.transforms.CreateNotebookSectionAnchors)
import json
import re

from panflute import Element, Doc, Span, Div, Math, Image, Table  # noqa: F401
import panflute as pf

from ipypublish.filters_pandoc.utils import convert_units, convert_attributes
from ipypublish.filters_pandoc.prepare_labels import (
    LABELLED_IMAGE_CLASS,
    LABELLED_MATH_CLASS,
    LABELLED_TABLE_CLASS,
)

try:
    from textwrap import indent
except ImportError:  # added in python 3.3

    def indent(text, prefix):
        return "".join(prefix + line for line in text.splitlines(True))


LATEX_FIG_LABELLED = """\\begin{{figure}}[{options}]
\\hypertarget{{{label}}}{{%
\\begin{{center}}
\\adjustimage{{max size={{0.9\\linewidth}}{{0.9\\paperheight}},{size}}}{{{path}}}
\\end{{center}}
\\caption{{{caption}}}\\label{{{label}}}
}}
\\end{{figure}}"""  # noqa: E501

LATEX_FIG_UNLABELLED = """\\begin{{figure}}[{options}]
\\begin{{center}}
\\adjustimage{{max size={{0.9\\linewidth}}{{0.9\\paperheight}},{size}}}{{{path}}}
\\end{{center}}
\\caption{{{caption}}}
\\end{{figure}}"""  # noqa: E501

MATH_ENVS = (
    "equation",
    "align",
    "alignat",
    "eqnarray",
    "multline",
    "gather",
    "flalign",
    "dmath",
)


def format_math(math, doc):
    # type: (Math, Doc) -> Element
    """
    originally adapted from:
    `pandoc-eqnos <https://github.com/tomduck/pandoc-eqnos/>`_
    """
    if not isinstance(math, pf.Math):
        return None

    if math.format != "DisplayMath":
        return None

    # test if the math text is already wrapped in an environment
    regex = re.compile(
        r"\\begin\{{((?:{0})\*?)\}}(.*)\\end\{{((?:{0})\*?)\}}".format(
            "|".join(MATH_ENVS)
        ),
        re.DOTALL,
    )
    wrap_match = regex.match(math.text)

    env = None
    label = None
    if isinstance(math.parent, pf.Span) and LABELLED_MATH_CLASS in math.parent.classes:
        span = math.parent
        numbered = "*" if "unnumbered" in span.classes else ""
        env = span.attributes.get("env", "equation") + numbered
        label = span.identifier

    if doc.format in ("tex", "latex"):
        if wrap_match:
            # TODO edge case where a label has been specified, but the math is already wrapped
            tex = math.text
        else:
            tex = "\\begin{{{0}}}{1}\\label{{{2}}}\\end{{{0}}}".format(
                env or "equation", math.text, label or ""
            )
        return pf.RawInline(tex, format="tex")

    elif doc.format in ("rst"):
        if env:
            tex = indent("\\begin{{{0}}}{1}\\end{{{0}}}".format(env, math.text), "   ")
        else:
            tex = indent(math.text.strip(), "   ")
        rst = "\n\n.. math::\n"
        if wrap_match or env:
            rst += "   :nowrap:\n"
        if label:
            rst += "   :label: {}\n".format(label)
        rst += "\n{}\n\n".format(tex)
        return pf.RawInline(rst, format="rst")

    elif doc.format in ("html", "html5"):
        # new_span = pf.Span(anchor_start, math, anchor_end)
        # TODO add formatting
        # TODO name by count
        if label:
            if not wrap_match:
                math.text = "\\begin{{{0}}}{1}\\end{{{0}}}".format(
                    env or "equation", math.text
                )
            return _wrap_in_anchor(math, label)
        else:
            return None


def format_image(image, doc):
    # type: (Image, Doc) -> Element
    """
    originally adapted from:
    `pandoc-fignos <https://github.com/tomduck/pandoc-fignos/>`_
    """
    if not isinstance(image, pf.Image):
        return None

    span = None
    if (
        isinstance(image.parent, pf.Span)
        and LABELLED_IMAGE_CLASS in image.parent.classes
    ):
        span = image.parent

    if span is not None:
        identifier = span.identifier
        attributes = span.attributes
        #  classes = span.classes
    else:
        identifier = image.identifier
        attributes = image.attributes
        # classes = image.classes

    if doc.format in ("tex", "latex"):
        new_doc = Doc(pf.Para(*image.content))
        new_doc.api_version = doc.api_version
        if image.content:
            caption = pf.run_pandoc(
                json.dumps(new_doc.to_json()), args=["-f", "json", "-t", "latex"]
            ).strip()
        else:
            caption = ""

        options = attributes.get("placement", "")
        size = ""  # max width set as 0.9\linewidth
        if "width" in attributes:
            width = convert_units(attributes["width"], "fraction")
            size = "width={0}\\linewidth".format(width)
        elif "height" in attributes:
            height = convert_units(attributes["height"], "fraction")
            size = "height={0}\\paperheight".format(height)

        if identifier:
            latex = LATEX_FIG_LABELLED.format(
                label=identifier,
                options=options,
                path=image.url,
                caption=caption,
                size=size,
            )
        else:
            latex = LATEX_FIG_UNLABELLED.format(
                options=options, path=image.url, caption=caption, size=size
            )

        return pf.RawInline(latex, format="tex")

    elif doc.format in ("rst",):
        if not image.content.list:
            # If the container is empty, then pandoc will assign an iterative
            # reference identifier to it (image0, image1).
            # However, this iterator restarts for each markdown cell,
            # which can lead to reference clashes.
            # Therefore we specifically assign the identifier here, as its url
            # TODO does this identifier need to be sanitized?
            # (it works fine in the tests)
            identifier = image.url
            image.content = pf.ListContainer(pf.Str(str(identifier)))

        return image
        # TODO formatting and span identifier (convert width/height to %)

    elif doc.format in ("html", "html5"):
        if identifier:
            return _wrap_in_anchor(image, identifier)
        else:
            return image
        # TODO formatting, name by count
    else:
        return None


def format_table(table, doc):
    # type: (Table, Doc) -> Element
    """
    originally adapted from:
    `pandoc-tablenos <https://github.com/tomduck/pandoc-tablenos>`_
    """
    if not isinstance(table, pf.Table):
        return None

    div = None  # type: pf.Div
    if (
        isinstance(table.parent, pf.Div)
        and LABELLED_TABLE_CLASS in table.parent.classes
    ):
        div = table.parent

    if div is None:
        return None

    attributes = convert_attributes(div.attributes)

    if "align" in div.attributes:
        align_text = attributes["align"]
        align = [
            {"l": "AlignLeft", "r": "AlignRight", "c": "AlignCenter"}.get(a, None)
            for a in align_text
        ]
        if None in align:
            raise ValueError(
                "table '{0}' alignment must contain only l,r,c:"
                " {1}".format(div.identifier, align_text)
            )
        table.alignment = align
        attributes["align"] = align

    if "widths" in div.attributes:
        widths = attributes["widths"]
        try:
            widths = [float(w) for w in widths]
        except Exception:
            raise ValueError(
                "table '{0}' widths must be a list of numbers:"
                " {1}".format(div.identifier, widths)
            )
        table.width = widths
        attributes["widths"] = widths

    if doc.format in ("tex", "latex"):
        # TODO placement
        table.caption.append(
            pf.RawInline("\\label{{{0}}}".format(div.identifier), format="tex")
        )
        return table

    if doc.format in ("rst",):
        # pandoc 2.6 doesn't output table options
        if attributes:
            tbl_doc = pf.Doc(table)
            tbl_doc.api_version = doc.api_version
            tbl_str = pf.convert_text(
                tbl_doc, input_format="panflute", output_format="rst"
            )

            tbl_lines = tbl_str.splitlines()
            if tbl_lines[1].strip() == "":
                tbl_lines.insert(1, "   :align: center")
                if "widths" in attributes:
                    # in rst widths must be integers
                    widths = " ".join([str(int(w * 10)) for w in table.width])
                    tbl_lines.insert(1, "   :widths: {}".format(widths))
            # TODO rst column alignment, see
            # https://cloud-sptheme.readthedocs.io/en/latest/lib/cloud_sptheme.ext.table_styling.html

            return [
                pf.Para(
                    pf.RawInline(".. _`{0}`:".format(div.identifier), format="rst")
                ),
                pf.RawBlock("\n".join(tbl_lines) + "\n\n", format=doc.format),
            ]

        return [
            pf.Para(pf.RawInline(".. _`{0}`:".format(div.identifier), format="rst")),
            table,
        ]

    if doc.format in ("html", "html5"):
        return _wrap_in_anchor(table, div.identifier, inline=False)
        # TODO formatting, name by count


def strip_labelled_spans(element, doc):
    # type: (Span, Doc) -> Element

    if isinstance(element, pf.Span) and set(element.classes).intersection(
        [LABELLED_IMAGE_CLASS, LABELLED_MATH_CLASS]
    ):
        return list(element.content)

    if isinstance(element, pf.Div) and set(element.classes).intersection(
        [LABELLED_TABLE_CLASS]
    ):
        return list(element.content)


def _wrap_in_anchor(element, label, inline=True):
    """ wrap element in html anchors

    according to https://stackoverflow.com/a/1828032/5033292
    can wrap inline and block elements in anchor in html5
    """
    if inline:
        raw = pf.RawInline
    else:
        raw = pf.RawBlock
    anchor_start = raw(
        '<a id="{0}" class="anchor-link" name="#{0}">'.format(label), format="html"
    )
    anchor_end = raw("</a>", format="html")
    return [anchor_start, element, anchor_end]


def prepare(doc):
    # type: (Doc) -> None
    pass


def finalize(doc):
    # type: (Doc) -> None
    pass


def main(doc=None, strip_spans=True):
    # type: (Doc, bool) -> None
    to_run = [format_math, format_image, format_table]
    if strip_spans:
        to_run.append(strip_labelled_spans)
    return pf.run_filters(to_run, prepare, finalize, doc=doc)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import logging
import sys

from ipypublish.frontend.shared import parse_options
from ipypublish.convert.main import IpyPubMain

logger = logging.getLogger("nbpublish")


def nbpublish(
    ipynb_path,
    outformat="latex_ipypublish_main",
    outpath=None,
    dump_files=True,
    ignore_prefix="_",
    clear_files=False,
    create_pdf=False,
    pdf_in_temp=False,
    pdf_debug=False,
    launch_browser=False,
    log_level="INFO",
    dry_run=False,
    print_traceback=False,
    export_paths=(),
):
    """ convert one or more Jupyter notebooks to a published format

    paths can be string of an existing file or folder,
    or a pathlib.Path like object

    Parameters
    ----------
    ipynb_path
        notebook file or directory
    outformat: str
        output format to use
    outpath : str or pathlib.Path
        path to output converted files
    dump_files: bool
        write files from nbconvert (containing images, etc) to outpath
    ignore_prefix: str
        ignore ipynb files with this prefix
    clear_files : str
        whether to clear existing external files in outpath folder
    create_pdf: bool
        convert to pdf (if converting to latex)
    pdf_in_temp: bool
        run pdf conversion in a temporary folder
        and only copy back the pdf file
    pdf_debug: bool
        run latexmk in interactive mode
    log_level: str
        the logging level (debug, info, critical, ...)

    """
    # run
    config = {
        "IpyPubMain": {
            "conversion": outformat,
            "plugin_folder_paths": export_paths,
            "outpath": outpath,
            "ignore_prefix": ignore_prefix,
            "log_to_stdout": True,
            "log_level_stdout": log_level,
            "log_to_file": True,
            "log_level_file": log_level,
            "default_pporder_kwargs": dict(
                dry_run=dry_run,
                clear_existing=clear_files,
                dump_files=dump_files,
                create_pdf=create_pdf,
            ),
            "default_ppconfig_kwargs": dict(
                pdf_in_temp=pdf_in_temp,
                pdf_debug=pdf_debug,
                launch_browser=launch_browser,
            ),
        }
    }
    publish = IpyPubMain(config=config)
    try:
        publish(ipynb_path)
    except Exception as err:
        logger.error("Run Failed: {}".format(err))
        if print_traceback:
            raise
        return 1

    return 0


def run(sys_args=None):

    if sys_args is None:
        sys_args = sys.argv[1:]

    filepath, options = parse_options(sys_args, "nbpublish")

    outcode = nbpublish(filepath, **options)

    return outcode



from sphinx.roles import XRefRole

from ipypublish.bib2glossary import BibGlossDB
from ipypublish.sphinx.gls.bibgloss import docutils_citation_ref_node


class GLSRole(XRefRole):
    """Class for processing the `gls` role."""

    def result_nodes(self, document, env, node, is_ref):
        """Transform reference node into a citation reference,
        and note that the reference was cited.
        """
        keys = node["reftarget"].split(",")
        # Note that at this point, usually, env.bibgloss_cache.bibfiles
        # is still empty because the bibliography directive may not
        # have been processed yet, so we cannot get the actual entry.
        # Instead, we simply fake an entry with the desired key, and
        # fix the label at doctree-resolved time. This happens in
        # process_citation_references.
        refnodes = [
            docutils_citation_ref_node(
                BibGlossDB.get_fake_entry_obj(key), document, classes=["bibglossary"]
            )
            for key in keys
        ]
        for key in keys:
            env.bibgloss_cache.add_cited(key, env.docname)
        return refnodes, []


class GLSCapitalRole(XRefRole):
    """Class for processing the `glsc` role."""

    def result_nodes(self, document, env, node, is_ref):
        """Transform reference node into a citation reference,
        and note that the reference was cited.
        """
        keys = node["reftarget"].split(",")
        # Note that at this point, usually, env.bibgloss_cache.bibfiles
        # is still empty because the bibliography directive may not
        # have been processed yet, so we cannot get the actual entry.
        # Instead, we simply fake an entry with the desired key, and
        # fix the label at doctree-resolved time. This happens in
        # process_citation_references.
        refnodes = [
            docutils_citation_ref_node(
                BibGlossDB.get_fake_entry_obj(key),
                document,
                classes=["bibglossary", "bibgcapital"],
            )
            for key in keys
        ]
        for key in keys:
            env.bibgloss_cache.add_cited(key, env.docname)
        return refnodes, []


class GLSPluralRole(XRefRole):
    """Class for processing the `glspl` role."""

    def result_nodes(self, document, env, node, is_ref):
        """Transform reference node into a citation reference,
        and note that the reference was cited.
        """
        keys = node["reftarget"].split(",")
        # Note that at this point, usually, env.bibgloss_cache.bibfiles
        # is still empty because the bibliography directive may not
        # have been processed yet, so we cannot get the actual entry.
        # Instead, we simply fake an entry with the desired key, and
        # fix the label at doctree-resolved time. This happens in
        # process_citation_references.
        refnodes = [
            docutils_citation_ref_node(
                BibGlossDB.get_fake_entry_obj(key),
                document,
                classes=["bibglossary", "bibgplural"],
            )
            for key in keys
        ]
        for key in keys:
            env.bibgloss_cache.add_cited(key, env.docname)
        return refnodes, []


class GLSPluralCapitalRole(XRefRole):
    """Class for processing the `glscpl` role."""

    def result_nodes(self, document, env, node, is_ref):
        """Transform reference node into a citation reference,
        and note that the reference was cited.
        """
        keys = node["reftarget"].split(",")
        # Note that at this point, usually, env.bibgloss_cache.bibfiles
        # is still empty because the bibliography directive may not
        # have been processed yet, so we cannot get the actual entry.
        # Instead, we simply fake an entry with the desired key, and
        # fix the label at doctree-resolved time. This happens in
        # process_citation_references.
        refnodes = [
            docutils_citation_ref_node(
                BibGlossDB.get_fake_entry_obj(key),
                document,
                classes=["bibglossary", "bibgplural", "bibgcapital"],
            )
            for key in keys
        ]
        for key in keys:
            env.bibgloss_cache.add_cited(key, env.docname)
        return refnodes, []



from docutils import nodes


class AdmonitionNode(nodes.Element):
    """A custom node for info and warning boxes."""


class CodeAreaNode(nodes.Element):
    """Input area or output area of a Jupyter notebook code cell."""


class FancyOutputNode(nodes.Element):
    """A custom node for non-code output of code cells."""
